import collections
import logging

import curio
import typing
from types import MappingProxyType

from curious.core import gateway
from curious.dataclasses.channel import Channel, ChannelType
from curious.dataclasses.emoji import Emoji
from curious.dataclasses.guild import Guild
from curious.dataclasses.member import Member
from curious.dataclasses.message import Message
from curious.dataclasses.permissions import Permissions
from curious.dataclasses.reaction import Reaction
from curious.dataclasses.role import Role
from curious.dataclasses.status import Game, Status
from curious.dataclasses.user import BotUser, User, FriendType
from curious.dataclasses.user import RelationshipUser
from curious.dataclasses.voice_state import VoiceState
from curious.dataclasses.webhook import Webhook

UserType = typing.TypeVar("U", bound=User)


class State(object):
    """
    This represents the state of the Client - in other libraries, the cache.

    It is also used to parse objects into their representations.
    """

    def __init__(self, client, max_messages: int = 500):
        #: The current user of this bot.
        #: This is automatically set after login.
        self._user = None  # type: User

        #: The client associated with this connection.
        self.client = client

        #: The private channel cache.
        self._private_channels = {}

        #: The guilds the bot can see.
        self._guilds = {}  # type: typing.Dict[int, Guild]

        #: This user's friends.
        self._friends = {}  # type: typing.Dict[int, RelationshipUser]

        #: This user's blocked users.
        self._blocked = {}  # type: typing.Dict[int, RelationshipUser]

        #: The current user cache.
        self._users = {}

        #: The state logger.
        self.logger = logging.getLogger("curious.state")

        #: The deque of messages.
        #: This is bounded to prevent the message cache from growing infinitely.
        self._messages = collections.deque(maxlen=max_messages)

        self.__shards_is_ready = collections.defaultdict(lambda *args, **kwargs: curio.Event())
        self.__voice_state_crap = collections.defaultdict(lambda *args, **kwargs: ((curio.Event(), curio.Event()), {}))

    def _is_ready(self, shard_id: int):
        return self.__shards_is_ready[shard_id]

    def _reset(self, shard_id: int):
        """
        Called after session is invalidated, to reset our state.
        """
        self.__shards_is_ready.pop(shard_id, None)

        for guild in self.guilds_for_shard(shard_id):
            guild._finished_chunking.clear()

    def _mark_for_chunking(self, gw: 'gateway.Gateway', guild: Guild):
        """
        Marks a guild for chunking.
        """
        if all(not g.unavailable for g in self.guilds_for_shard(guild.shard_id)):
            gw._enqueued_guilds.append(guild)
            # we have all guilds anyway, so raise ChunkGuilds NOW
            raise gateway.ChunkGuilds

        if len(gw._enqueued_guilds) >= 74:
            # bump it up to 74
            gw._enqueued_guilds.append(guild)
            raise gateway.ChunkGuilds

        gw._enqueued_guilds.append(guild)

    @property
    def guilds(self) -> typing.Mapping[int, Guild]:
        return MappingProxyType(self._guilds)

    def have_all_chunks(self, shard_id: int):
        """
        Checks if we have all the chunks for the specified shard.

        :param shard_id: The shard ID to check.
        """
        return all(guild._finished_chunking.is_set()
                   for guild in self.guilds.values()
                   if guild.shard_id == shard_id)

    def guilds_for_shard(self, shard_id: int):
        """
        Gets all the guilds for a particular shard.
        """
        return [guild for guild in self.guilds.values() if guild.shard_id == shard_id]

    async def _check_ready(self, gw: 'gateway.Gateway', guild: Guild):
        """
        Checks if we should dispatch ready for this shard.
        """
        if self._is_ready(gw.shard_id).is_set():
            # Already ready, don't bother.
            return

        if not self.client.is_bot:
            # User chunk handling is setup differently.
            return

        if self.have_all_chunks(gw.shard_id):
            # Have all chunks anyway, dispatch now.
            self.logger.info("All guilds fully chunked on shard {}, dispatching READY.".format(gw.shard_id))
            await self.client.fire_event("ready", gateway=gw)
            await self._is_ready(gw.shard_id).set()

        # Check if we need to forcibly fire ChunkGuilds.
        if all(g.unavailable is False for g in self.guilds_for_shard(gw.shard_id)):
            raise gateway.ChunkGuilds

    # get_all_* methods
    def get_all_channels(self):
        for guild in self._guilds.values():
            for channel in guild.channels.values():
                yield channel

    def get_all_members(self):
        for guild in self.guilds.values():
            for member in guild.members.values():
                yield member

    def get_all_roles(self):
        for guild in self.guilds.values():
            for role in guild.roles.values():
                yield role

    def _get_channel(self, channel_id: int) -> Channel:
        """
        Internal method to get a channel from any guild.
        """
        # default channel_id == guild id
        if channel_id in self._guilds:
            try:
                return self._guilds[channel_id].default_channel
            except KeyError:
                return None

        if channel_id in self._private_channels:
            return self._private_channels[channel_id]

        for channel in self.get_all_channels():
            if channel.id == channel_id:
                return channel

    def _find_message(self, message_id: int) -> Message:
        for message in reversed(self._messages):
            if message.id == message_id:
                return message

    def _check_decache_user(self, id: int):
        """
        Checks if we should decache a user.

        This will check if there is any guild with a reference to the user.
        """
        if id in self._friends or id in self._blocked:
            return

        for guild in self._guilds.values():
            if id in guild.members:
                return

        # didn't return, so no references
        self._users.pop(id, None)

    def make_webhook(self, event_data: dict) -> Webhook:
        """
        Creates a new webhook object from the event data.

        :param event_data: The event data.
        :return: A :class:`Webhook`.
        """
        if "content" in event_data:
            # message object, so we have to do a minor bit of remapping
            user = event_data.get("author", {})
            webhook_id = int(event_data.pop("webhook_id", 0))
            owner = {}
        else:
            # make a "fake" user
            webhook_id = int(event_data.pop("id", 0))
            user = {
                "id": webhook_id,
                "discriminator": "0000",
                "avatar": event_data.get("avatar"),
                "username": event_data.get("username")
            }
            owner = event_data.get("user", {})

        channel = self._get_channel(int(event_data.get("channel_id")))
        user = User(self.client, **user)
        user.bot = True
        webhook = Webhook(client=self.client, webhook_id=webhook_id, **event_data)
        webhook.channel = channel
        webhook.user = user
        webhook.token = event_data.get("token", None)

        if owner:
            # only create Owner if the data was returned
            webhook.owner = self.make_user(owner)

        # default fields, these are lazily loaded by properties
        webhook._default_name = event_data.get("name", None)
        webhook._default_avatar = event_data.get("avatar", None)

        return webhook

    def make_private_channel(self, channel_data: dict) -> Channel:
        """
        Creates a new private channel and caches it.

        :param channel_data: The channel data to cache.
        """
        channel = Channel(self.client, guild=None, **channel_data)
        self._private_channels[channel.id] = channel

        return channel

    def make_user(self, user_data: dict, *,
                  user_klass: typing.Type[UserType] = User,
                  override_cache: bool=False) -> UserType:
        """
        Creates a new user and caches it.

        :param user_data: The user data to create.
        :param user_klass: The type of user to create.
        :param override_cache: Should the cache be overridden?
        """
        id = int(user_data.get("id", 0))
        if id in self._users and not override_cache:
            return self._users[id]

        user = user_klass(self.client, **user_data)
        self._users[user.id] = user

        return user

    async def wait_for_voice_data(self, guild_id: int):
        """
        Waits for the two voice data packets to be received for the specified guild.
        """
        events, state = self.__voice_state_crap[guild_id]  # state is a pointer

        for event in events:
            await event.wait()

        # pop out the event data for laeter reconnects
        self.__voice_state_crap.pop(guild_id)

        return state

    # Event handlers.
    # These parse the events and deconstruct them.
    async def handle_ready(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when READY is dispatched.
        """
        gw.session_id = event_data.get("session_id")

        # Create our bot user.
        self._user = BotUser(self.client, **event_data.get("user"))
        # cache ourselves
        self._users[self._user.id] = self._user

        self.logger.info("Parsing ready for `{}#{}` ({})".format(self._user.username, self._user.discriminator,
                                                                 self._user.id))

        self.logger.info("Logged in as a userbot: {}".format(not self._user.bot))

        # User-bots only.
        # Parse relationships first before processing anything else.
        # This ensures all friends are calculated and assigned.
        if not self._user.bot:
            # Parse friends and blocked users.
            for item in event_data.get("relationships", []):
                # make user to cache it
                user = self.make_user(item.get("user", {}), user_klass=RelationshipUser)
                user.type_ = FriendType(item["type"])
                if user.type_ == FriendType.FRIEND:
                    self._friends[int(item["id"])] = user
                else:
                    self._blocked[int(item["id"])] = user

            for presence in event_data.get("presences", []):
                u = int(presence["user"]["id"])
                fr = self._friends.get(u)

                # eventual consistency
                if not fr:
                    continue

                fr.status = Status(presence["status"])
                gm = presence["game"]
                if gm is not None:
                    fr.game = Game(**gm)

            self.logger.info("Processed {} friends "
                             "and {} blocked users.".format(len(self._friends), len(self._blocked)))

        # Create all of the private channels.
        for channel in event_data.get("private_channels"):
            self.make_private_channel(channel)

        self.logger.info("Processed {} private channels.".format(len(self._private_channels)))

        # Create all of the guilds.
        for guild in event_data.get("guilds"):
            new_guild = Guild(self.client, **guild)
            new_guild.shard_id = gw.shard_id
            self._guilds[new_guild.id] = new_guild

            if not self._user.bot and len(event_data.get("guilds")) <= 100:
                # this might as well be a GUILD_CREATE, so treat it as one
                await self.handle_guild_create(gw, guild)

        if not self._user.bot and len(event_data.get("guilds")) <= 100:
            await gw.send_guild_sync([g.id for g in self.guilds.values()])
            self.logger.info("Syncing {} guilds.".format(len(self.guilds)))

        self.logger.info("Ready processed for shard {}. Delaying until all guilds are chunked.".format(gw.shard_id))

        # Don't fire `_ready` here, because we don't have all guilds.
        await self.client.fire_event("connect", gateway=gw)

        # However, if the client has no guilds, we DO want to fire ready.
        if len(event_data.get("guilds", {})) == 0 or not self._user.bot:
            await self._is_ready(gw.shard_id).set()
            self.logger.info("No more guilds to get for shard {}, or client is user. "
                             "Dispatching READY.".format(gw.shard_id))
            await self.client.fire_event("ready", gateway=gw)

    async def handle_resumed(self, gw: 'gateway.Gateway', event_data: dict):
        self.logger.info("Successfully resumed session on shard ID {}.".format(gw.shard_id))
        await self.client.fire_event("resumed", gateway=gw)

    async def handle_user_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when the bot's user is updated.
        """
        id = event_data.get("id")

        self._user.id = int(id)
        self._user.username = event_data.get("username")
        self._user.discriminator = event_data.get("discriminator")
        self._user._avatar_hash = event_data.get("avatar")

        await self.client.fire_event("user_update", gateway=gw)

    async def handle_presence_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a member changes game.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            # user presence update
            user_id = int(event_data["user"]["id"])
            fr = self._friends.get(user_id)

            if not fr:
                # wtf
                return

            # re-create the user object
            self._friends[user_id] = self.make_user(event_data["user"], user_klass=RelationshipUser,
                                                    override_cache=True)

            fr.status = Status(event_data.get("status"))
            game = event_data.get("game")
            if game is None:
                fr.game = None
            else:
                fr.game = Game(**game)

            await self.client.fire_event("friend_update", fr, gateway=gw)
            return

        member_id = int(event_data["user"]["id"])
        member = guild.members.get(member_id)

        if member is None:
            # thanks discord
            return

        old_member = member._copy()

        game = event_data.get("game", {})
        if game is not None:
            member.game = Game(**game, status=event_data.get("status"))
        else:
            member.game = None

        roles = event_data.get("roles", [])
        for role_id in roles:
            role_id = int(role_id)
            role = guild.roles.get(role_id)

            if not role:
                continue

            member._roles[role.id] = role

        member.status = event_data.get("status")

        if not isinstance(member.user, RelationshipUser):
            # recreate the user object
            self.make_user(event_data["user"], override_cache=True)

        await self.client.fire_event("member_update", old_member, member, gateway=gw)

    async def handle_guild_members_chunk(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a chunk of members has arrived.
        """
        id = int(event_data.get("guild_id"))
        guild = self._guilds.get(id)

        if not guild:
            self.logger.warning("Got a chunk for a Guild that doesn't exist...")
            return

        members = event_data.get("members", [])
        self.logger.info("Got a chunk of {} members in guild {} "
                         "on shard {}".format(len(members), guild.name or guild.id, guild.shard_id))

        guild._handle_member_chunk(event_data.get("members"))
        await self.client.fire_event("guild_chunk", guild, gateway=gw)

        if guild._chunks_left <= 0:
            # Set the finished chunking event.
            await guild._finished_chunking.set()
            await self.client.fire_event("guild_available", guild, gateway=gw)

        # Check if we have all chunks.
        await self._check_ready(gw, guild)

    async def handle_guild_sync(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a guild is synced.
        """
        id = int(event_data.get("id", 0))
        guild = self._guilds.get(id)

        if not guild:
            # why
            self.logger.warning("Got sync for guild we're not in...")
            return

        members = event_data.get("members", [])
        # same logic as a guild chunk
        guild._handle_member_chunk(members)

        presences = event_data.get("presences", [])

        for presence in presences:
            u_id = presence["user"]["id"]
            member = guild.members.get(int(u_id))

            if not member:
                continue

            game = presence.get("game", None)
            if game is None:
                game = {}

            member.game = Game(**game)
            member.status = presence.get("status")

        self.logger.info("Processed a guild sync for guild {} with "
                         "{} members and {} presences.".format(guild.name, len(members), len(presences)))

    async def handle_guild_create(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when GUILD_CREATE is dispatched.
        """
        id = int(event_data.get("id", 0))
        guild = self._guilds.get(id)

        had_guild = True
        if guild:
            guild.from_guild_create(**event_data)
        else:
            had_guild = False
            guild = Guild(self.client, **event_data)
            self._guilds[guild.id] = guild

        guild.shard_id = gw.shard_id
        guild.me.game = gw.game
        guild.me.status = gw.status

        # Dispatch the event if we're ready (i.e not streaming)
        if self._is_ready(gw.shard_id).is_set():
            if had_guild:
                await self.client.fire_event("guild_available", guild, gateway=gw)
            else:
                # We didn't have it before, so we just joined it.
                # Hence, we fire a `guild_join` event.
                await self.client.fire_event("guild_join", guild, gateway=gw)
        else:
            self.logger.debug("Streamed guild: {} ({})".format(guild.name, guild.id))
            await self.client.fire_event("guild_streamed", guild, gateway=gw)

        if not guild.unavailable and guild.large:
            # mark this guild as a chunking guild
            self._mark_for_chunking(gw, guild)
        else:
            # set finished_chunking now
            await guild._finished_chunking.set()

        await self._check_ready(gw, guild)

    async def handle_guild_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when GUILD_UPDATE is dispatched.
        """
        id = int(event_data.get("id", 0))
        guild = self._guilds.get(id)

        if not guild:
            return

        old_guild = guild._copy()

        guild.unavailable = event_data.get("unavailable", False)
        guild.name = event_data.get("name")
        guild.member_count = event_data.get("member_count", 0)
        if not guild.member_count:
            guild.member_count = len(guild._members)
        guild.large = event_data.get("large")
        guild._icon_hash = event_data.get("icon")
        guild._splash_hash = event_data.get("splash")
        guild.region = event_data.get("region")
        guild.mfa_level = event_data.get("mfa_level")
        guild._afk_channel_id = int(event_data.get("afk_channel", 0))
        guild.afk_timeout = event_data.get("afk_timeout")
        guild.verification_level = event_data.get("verification_level")
        guild._owner_id = int(event_data.get("owner_id", 0))

        await self.client.fire_event("guild_update", old_guild, guild, gateway=gw)

    async def handle_guild_delete(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a guild becomes unavailable.
        """
        guild_id = int(event_data.get("id", 0))
        # Check if the `unavailable` flag is there.
        # If it is, we want to semi-discard this event, because all it means is the guild becomes unavailable.
        if event_data.get("unavailable", False):
            # Set the guild to unavailable, but don't delete it.
            guild = self._guilds.get(guild_id)
            if guild:
                guild.unavailable = True
                await self.client.fire_event("guild_unavailable", guild, gateway=gw)

        else:
            # We've left this guild - clear it from our dictionary of guilds.
            guild = self._guilds.pop(guild_id, None)
            if guild:
                await self.client.fire_event("guild_leave", guild, gateway=gw)

    async def handle_guild_emojis_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a guild updates its emojis.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        old_guild = guild._copy()
        emojis = event_data.get("emojis", [])
        guild._handle_emojis(emojis)

        await self.client.fire_event("guild_emojis_update", old_guild, guild, gateway=gw)

    def parse_message(self, event_data: dict, cache: bool = True) -> Message:
        message = Message(self.client, **event_data)
        # discord won't give us the Guild id
        # so we have to search it from the channels
        channel_id = int(event_data.get("channel_id"))
        channel = self._get_channel(channel_id)
        if not channel:
            # fuck off discord
            return None

        author_id = int(event_data.get("author", {}).get("id", 0))

        message.channel = channel
        message.guild = channel.guild
        if message.channel.type == ChannelType.PRIVATE:
            message.author = self._user
        elif message.channel.type == ChannelType.GROUP:
            message.author = next(filter(lambda m: m.id == author_id, message.channel.recipients), None)
        else:
            # Webhooks also exist.
            if event_data.get("webhook_id") is not None:
                message.author = self.make_webhook(event_data)
            else:
                message.author = message.guild.members.get(author_id)

        for reaction_data in event_data.get("reactions", []):
            emoji = reaction_data.pop("emoji")
            reaction = Reaction(**reaction_data)

            if "id" in emoji and emoji["id"] is not None:
                emoji_obb = message.guild.emojis.get(int(emoji["id"]))
                if emoji_obb is None:
                    emoji_obb = Emoji(id=emoji["id"], name=emoji["name"])
            else:
                emoji_obb = emoji.get("name", None)

            reaction.emoji = emoji_obb
            message.reactions.append(reaction)

        if cache and message not in self._messages:
            self._messages.append(message)

        return message

    async def handle_message_create(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when MESSAGE_CREATE is dispatched.
        """
        message = self.parse_message(event_data)
        if not message:
            return

        # Hope that messages are ordered!
        message.channel._last_message_id = message.id

        await self.client.fire_event("message_create", message, gateway=gw)

    async def handle_message_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when MESSAGE_UPDATE is dispatched.
        """
        new_message = self.parse_message(event_data, cache=False)
        if not new_message:
            return

        # Try and find the old message.
        old_message = self._find_message(new_message.id)
        if not old_message:
            return

        self._messages.remove(old_message)
        self._messages.append(new_message)

        if old_message.content != new_message.content:
            # Fire a message_edit, as well as a message_update, because the content differs.
            await self.client.fire_event("message_edit", old_message, new_message, gateway=gw)

        await self.client.fire_event("message_update", old_message, new_message, gateway=gw)

    async def handle_message_delete(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when MESSAGE_DELETE is dispatched.
        """
        message_id = int(event_data.get("id"))
        message = self._find_message(message_id)

        if not message:
            return

        await self.client.fire_event("message_delete", message, gateway=gw)

    async def handle_message_delete_bulk(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when MESSAGE_DELETE_BULK is dispatched.
        """
        messages = []
        for message in event_data.get("ids", []):
            message = self._find_message(int(message))
            if not message:
                continue

            messages.append(message)

        await self.client.fire_event("message_delete_bulk", messages, gateway=gw)

    def _find_emoji(self, emoji_data: dict):
        if emoji_data.get("id", None) is None:
            # str only
            return emoji_data["name"]

        # try and get it from the guilds
        for guild in self.guilds.values():
            em = guild.emojis.get(int(emoji_data["id"]))
            if em:
                return em

    async def handle_message_reaction_add(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a reaction is added to a message.
        """
        message_id = event_data.get("message_id")
        if not message_id:
            return

        message_id = int(message_id)
        message = self._find_message(message_id)

        if not message:
            return

        # complex filter
        def _f(r: Reaction):
            if not r.emoji:
                return False

            e = self._find_emoji(event_data["emoji"])
            if not e:
                # ¯\_(ツ)_/¯
                return False

            return r.emoji == e

        reaction = next(filter(_f, message.reactions), None)

        if not reaction:
            emoji = event_data.pop("emoji", {})
            # no useful args are added
            reaction = Reaction()

            if "id" in emoji and emoji["id"] is not None:
                emoji_obb = message.guild.emojis.get(int(emoji["id"]))
                if emoji_obb is None:
                    emoji_obb = Emoji(id=emoji["id"], name=emoji["name"])
            else:
                emoji_obb = emoji.get("name", None)

            reaction.emoji = emoji_obb
            message.reactions.append(reaction)
        else:
            # up the count
            reaction.count += 1
            # if our user id matches, we obviously voted on it
            if int(event_data["user_id"]) == self._user.id:
                reaction.me = True

        member_id = int(event_data.get("user_id", 0))
        channel = self._get_channel(int(event_data.get("channel_id", 0)))
        if channel.guild:
            author = channel.guild.members.get(member_id)
        else:
            author = channel.user

        await self.client.fire_event("message_reaction_add", message, author, reaction, gateway=gw)

    async def handle_message_reaction_remove_all(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when all reactions are removed from a message.
        """
        message = self._find_message(int(event_data.get("message_id", 0)))
        if not message:
            return

        reactions = message.reactions.copy()
        message.reactions = []
        await self.client.fire_event("message_reaction_remove_all", message, reactions, gateway=gw)

    async def handle_message_reaction_remove(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a reaction is removed from a message.
        """
        message_id = event_data.get("message_id")
        if not message_id:
            return

        message_id = int(message_id)
        message = self._find_message(message_id)

        if not message:
            return

        def _f(r: Reaction):
            if not r.emoji:
                return False

            e = self._find_emoji(event_data["emoji"])
            if not e:
                # ¯\_(ツ)_/¯
                return False

        reaction = next(filter(_f, message.reactions), None)
        if not reaction:
            # nothing to do
            return

        reaction.count -= 1
        if int(event_data["user_id"]) == self._user.id:
            reaction.me = False

        if reaction.count == 0:
            message.reactions.remove(reaction)

        await self.client.fire_event("message_reaction_remove", message, reaction, gateway=gw)

    async def handle_guild_member_add(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a guild adds a new member.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        member = Member(self.client, **event_data)
        member.guild = guild

        guild._members[member.id] = member
        guild.member_count += 1
        await self.client.fire_event("member_join", member, gateway=gw)

    async def handle_guild_member_remove(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a guild removes a member.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        member = guild._members.pop(int(event_data["user"]["id"]), None)
        guild.member_count -= 1
        if not member:
            # We can't see the member, so don't fire an event for it.
            return

        # check if we should decache the user
        self._check_decache_user(member.id)

        await self.client.fire_event("member_leave", member, gateway=gw)

    async def handle_guild_member_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a guild member is updated.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        member_id = int(event_data["user"]["id"])
        member = guild.members.get(member_id)

        if not member:
            return

        # Make a copy of the member for the old previous reference.
        old_member = member._copy()
        # Re-create the user object.
        #self.make_user(event_data["user"], override_cache=True)
        #self._users[member.user.id] = member.user

        # Overwrite roles, we want to get rid of any roles that are stale.
        member._roles = {}

        for role_id in event_data.get("roles", []):
            role_id = int(role_id)
            role = guild.roles.get(role_id)

            if not role:
                # thanks discord
                continue

            member._roles[role.id] = role

        guild._members[member.id] = member

        member.nickname = event_data.get("nickname")
        await self.client.fire_event("member_update", old_member, member, gateway=gw)

    async def handle_guild_ban_add(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a ban is added to a guild.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if guild is None:
            return

        member_id = int(event_data["user"]["id"])
        member = guild.members.get(member_id)

        if not member:
            # Dispatch to `user_ban` instead of `member_ban`.
            user = self.make_user(event_data["user"])
            await self.client.fire_event("user_ban", guild, user, gateway=gw)
            return

        await self.client.fire_event("member_ban", member, gateway=gw)

    async def handle_guild_ban_remove(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a ban is removed from a guild.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if guild is None:
            return

        user = self.make_user(**event_data["user"])
        await self.client.fire_event("user_unban", guild, user, gateway=gw)

    async def handle_channel_create(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a channel is created.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        channel = Channel(self.client, guild=guild, **event_data)
        if channel.is_private:
            self._private_channels[channel.id] = channel
        else:
            if channel.id not in guild._channels:
                guild._channels[channel.id] = channel
            else:
                channel = guild._channels[channel.id]

        await self.client.fire_event("channel_create", channel, gateway=gw)

    async def handle_channel_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a channel is updated.
        """
        channel_id = int(event_data.get("id"))
        channel = self._get_channel(channel_id)

        if not channel:
            return

        old_channel = channel._copy()

        channel.name = event_data.get("name", channel.name)
        channel.position = event_data.get("position", channel.position)
        channel.topic = event_data.get("topic", channel.topic)

        channel._update_overwrites(event_data.get("permission_overwrites", []))
        await self.client.fire_event("channel_update", old_channel, channel, gateway=gw)

    async def handle_channel_delete(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a channel is deleted.
        """
        channel_id = int(event_data.get("channel_id", 0))
        channel = self._get_channel(channel_id)

        if not channel:
            return

        if channel.is_private:
            del self._private_channels[channel.id]
        else:
            del channel.guild._channels[channel.id]

        await self.client.fire_event("channel_delete", channel, gateway=gw)

    async def handle_guild_role_create(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a role is created.
        """
        guild_id = int(event_data.get("guild_id"))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        if int(event_data.get("role", {}).get("id", 0)) not in guild._roles:
            role = Role(self.client, **event_data.get("role", {}))
            guild._roles[role.id] = role
        else:
            role = guild._roles[event_data["role"].get("id", 0)]

        await self.client.fire_event("role_create", role, gateway=gw)

    async def handle_guild_role_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a role is updated.
        """
        guild_id = int(event_data.get("guild_id"))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        role = guild.roles.get(int(event_data["role"]["id"]))

        if not role:
            return

        old_role = role._copy()

        # Update all the fields on the role.
        event_data = event_data.get("role", {})
        role.colour = event_data.get("color", 0)
        role.name = event_data.get("name")
        role.position = event_data.get("position")
        role.hoisted = event_data.get("hoisted")
        role.mentionable = event_data.get("mentionable")
        role.managed = event_data.get("managed")
        role.permissions = Permissions(event_data.get("permissions", 0))

        await self.client.fire_event("role_update", old_role, role, gateway=gw)

    async def handle_guild_role_delete(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a role is deleted.
        """
        guild_id = int(event_data.get("guild_id"))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        role = guild._roles.pop(int(event_data["role_id"]), None)

        if not role:
            return

        # Remove the role from all members.
        for member in guild.members:
            member._roles.pop(role.id, None)

        await self.client.fire_event("role_delete", role, gateway=gw)

    async def handle_typing_start(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a user starts typing.
        """
        member_id = int(event_data.get("user_id"))
        channel_id = int(event_data.get("channel_id"))

        channel = self._get_channel(channel_id)
        if not channel:
            return

        if not channel.is_private:
            member = channel.guild.members.get(member_id)
            if not member:
                return
        else:
            member = channel.user

        await self.client.fire_event("user_typing", channel, member, gateway=gw)

    # Voice bullshit
    async def handle_voice_server_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a voice server update packet is dispatched.
        """
        # This is an internal event, so we do NOT dispatch it.
        # Instead, store some state internally.
        guild_id = event_data.get("guild_id")
        if not guild_id:
            # thanks
            return

        events, state = self.__voice_state_crap[int(guild_id)]

        state.update({
            "token": event_data.get("token"),
            "endpoint": event_data.get("endpoint"),
        })

        # Set the VOICE_SERVER_UPDATE event.
        await events[0].set()

    async def handle_voice_state_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a member's voice state changes.
        """
        guild_id = int(event_data.get("guild_id", 0))

        # If user_id == self._user.id, it's voice conn bullshit
        user_id = int(event_data.get("user_id", 0))
        if user_id == self._user.id:
            # YAY
            events, state = self.__voice_state_crap[guild_id]
            state.update({
                "session_id": event_data.get("session_id")
            })
            # 2nd one is voice_state_update event
            await events[1].set()
            return

        # get the guild and member
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        member = guild.members.get(user_id)
        if not member:
            return

        channel_id = event_data.get("channel_id", 0)
        if channel_id is None:
            # disconnect!
            new_voice_state = None
        else:
            new_voice_state = VoiceState(member.user, **event_data)
            new_voice_state.guild = guild
            new_voice_state.channel = guild.channels.get(new_voice_state._channel_id)

        old_voice_state = member.voice

        member.voice = new_voice_state

        await self.client.fire_event("voice_state_update", member, old_voice_state, new_voice_state, gateway=gw)

    async def handle_webhooks_update(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a channel has a webhook updated.

        This event is effectively useless.
        """

    # Userbot only events.
    async def handle_message_ack(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a message is acknowledged.
        """
        channel = self._get_channel(int(event_data.get("channel_id", 0)))
        message = self._find_message(int(event_data.get("message_id", 0)))

        if channel is None:
            return

        if message is None:
            return

        await self.client.fire_event("message_ack", channel, message, gateway=gw)

    async def handle_relationship_add(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a relationship is added.
        """
        type_ = event_data.get("type", 0)

        if type_ == 1:
            # FR accepted
            u = RelationshipUser(client=self.client, **event_data.get("user"))
            u.type_ = FriendType.FRIEND
            self._friends[u.id] = u
        elif type_ == 2:
            u = RelationshipUser(client=self.client, **event_data.get("user"))
            u.type_ = FriendType.BLOCKED
            self._blocked[u.id] = u
        else:
            return

        self._users[u.id] = u

        await self.client.fire_event("relationship_add", u, gateway=gw)

    async def handle_relationship_remove(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a relationship is removed.
        """
        type_ = event_data.get("type", 0)
        u_id = int(event_data.get("id", 0))

        if type_ == 1:
            u = self._friends.pop(u_id, None)
        elif type_ == 2:
            u = self._blocked.pop(u_id, None)
        else:
            return

        # replace the type in `_users` with a generic user
        # this wont cause race conditions due to the GIL etc
        del self._users[u_id]
        new_user = self.make_user(
            {
                "id": u_id,
                "username": u.username,
                "discriminator": u.discriminator,
                "avatar": u._avatar_hash,
                "bot": u.bot
            }
        )

        # maybe decache it anyway
        self._check_decache_user(u_id)

        await self.client.fire_event("relationship_remove", u, gateway=gw)

    async def handle_channel_recipient_add(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a recipient is added to a channel.
        """
        user = event_data.get("user", {})
        id = int(event_data.get("channel_id", 0))

        user = self._users.get(int(user.get("id", 0)))

        channel = self._get_channel(channel_id=id)
        if channel is None:
            return

        channel.recipients.append(user)

        await self.client.fire_event("group_user_add", channel, user, gateway=gw)

    async def handle_channel_recipient_remove(self, gw: 'gateway.Gateway', event_data: dict):
        """
        Called when a recipient is removed a channel.
        """
        user = event_data.get("user", {})
        id = int(event_data.get("channel_id", 0))

        user = self._users.get(int(user.get("id", 0)))
        channel = self._get_channel(channel_id=id)
        if channel is None:
            return

        if user in channel.recipients:
            channel.recipients.remove(user)
            await self.client.fire_event("group_user_remove", channel, user, gateway=gw)



"""
Represents a voice client.

Other documentation:

- Timestamp increases by 48 * 20 every time (samples * frame length).
- Sequence and timestamp both overflow.

frame_length = 20 (??)
sample_size = 4 (2 * 2)
samples_per_frame: 48 * 20
frame_size = samples_per_frame APPARENTLY


"""
import logging
import socket
import struct

import curio
from nacl.secret import SecretBox
from opuslib import Encoder

from curious.dataclasses import channel as dt_channel
from curious.voice import voice_player as vp
from curious.voice.voice_gateway import VoiceGateway

logger = logging.getLogger("curious.voice")


def simulate_overflow(value: int, bits: int, signed: bool):
    """
    Simulates integer overflow.
    """
    base = 1 << bits
    value %= base
    return value - base if signed and value.bit_length() == bits else value


class VoiceClient(object):
    """
    The voice client instance controls connecting to Discord's voice servers.

    This should ***not*** be created directly - instead use :class:`Channel.connect()` to connect to a voice channel,
    and use the instance returned from.
    """

    def __init__(self, main_client: 'client.Client',
                 channel: 'dt_channel.Channel'):
        #: The main client this voice client is associated with.
        self.client = main_client

        #: The voice channel this client is connected to.
        self.channel = channel

        #: The voice websocket that we are connected to.
        self.vs_ws = None  # type: VoiceGateway

        #: The UDP socket that we are connected to.
        self._sock = None  # type: socket.socket

        self.main_task = None  # type: curio.Task

        # Header related stuff
        self.sequence = 0  # sequence is 2 bytes
        self.timestamp = 0  # timestamp is 4 bytes

        self.encoder = Encoder(48000, 2, 'audio')
        self.encoder.bitrate = 96000

    @property
    def open(self):
        return self.vs_ws._open

    # Voice encoder related things.
    def _get_packet_header(self):
        header = bytearray(12)

        # constant values, provided by the docs
        header[0:2] = b"\x80\x78"

        # dynamic values
        # offset 2 -> sequence
        struct.pack_into(">H", header, 2, self.sequence)
        # offset 4 -> timestamp
        struct.pack_into(">I", header, 4, self.timestamp)
        # offset 8 -> ssrc
        struct.pack_into(">I", header, 8, self.vs_ws.ssrc)

        return header

    def _get_voice_packet(self, opus_body: bytes):
        header = self._get_packet_header()
        nonce = bytearray(24)
        # copy the header into nonce
        nonce[:12] = header

        encryptor = SecretBox(bytes(self.vs_ws.secret_key))
        encrypted_body = encryptor.encrypt(opus_body, nonce=bytes(nonce))

        # pack_nonce is True, so the body can just be concatted
        return bytes(header + encrypted_body.ciphertext)

    def _get_ip_discovery_packet(self):
        packet = bytearray(70)
        packet[0:4] = struct.pack(">I", self.vs_ws.ssrc)

        return packet

    def _send_voice_packet(self, built_packet: bytes):
        """
        Sends a voice packet.

        :param built_packet: The final built packet, as got by `_get_voice_packet`.
        """
        # Overflow values as appropriate.
        sequence = self.sequence + 1
        self.sequence = simulate_overflow(sequence, 2 * 8, False)

        timestamp = self.timestamp + (20 * 48)
        self.timestamp = simulate_overflow(timestamp, 2 * 16, False)

        try:
            self._sock.sendto(built_packet, (self.vs_ws.endpoint, self.vs_ws.port))
        except BlockingIOError:
            # can't send rn, oh well
            # opus is built for this
            logger.error("Failed to send voice packet! Sequence: {}, timestamp: {}".format(self.sequence,
                                                                                           self.timestamp))

    def send_opus_data(self, opus_data: bytes):
        """
        Sends an opus-encoded voice packet.

        :param opus_data: The data to send.
        """
        packet = self._get_voice_packet(opus_data)
        self._send_voice_packet(packet)

    def send_voice_packet(self, voice_data: bytes):
        """
        Sends voice data from a PCM buffer.

        :param voice_data: The raw PCM voice data to send.
        """
        data = self.encoder.encode(voice_data, 960)
        self.send_opus_data(data)

    def play_path(self, path: str):
        """
        Plays a file using ffmpeg.
        """
        player = vp.VoicePlayer(self, path)
        player.start()

    async def poll(self):
        """
        Polls the voice websocket constantly for new events.
        """
        while True:
            await self.vs_ws.next_event()

    async def close(self):
        await self.vs_ws._close()
        await self.main_task.cancel()

    async def connect(self, timeout=10):
        """
        Connects the voice client to the UDP server.
        """
        logger.info("Opening UDP connection to Discord voice servers.")
        # Wait before the voice socket is ready.
        await self.vs_ws._ready.wait()
        # Open our UDP connection.
        # TODO: Make this IPv6 compatible, if appropriate.
        # Now, you may be wondering why we use raw UDP sockets instead of using "nice" sockets.
        # This is because we need a voice thread to be able to access it.
        # This prevents blocking actions from killing the entire voice connection.
        addrinfo = await curio.abide(socket.getaddrinfo, self.vs_ws.endpoint, self.vs_ws.port,
                                     type=socket.SOCK_DGRAM)

        for item in addrinfo:
            try:
                new_socket = socket.socket(item[0], item[1])
                break
            except:
                continue
        else:
            raise ConnectionError("Could not create voice socket")

        self._sock = new_socket

        # Send an IP discovery packet.
        logger.info("Connecting to {}:{}".format(self.vs_ws.endpoint, self.vs_ws.port))
        packet = self._get_ip_discovery_packet()
        logger.debug("Sending IP discovery packet")
        self._sock.sendto(packet, (self.vs_ws.endpoint, self.vs_ws.port))

        # Wait for our local IP to be received.
        try:
            packet_data, addr = await curio.timeout_after(timeout, curio.abide(self._sock.recvfrom, 70))
        except curio.TaskTimeout:
            self._sock.close()
            self.vs_ws._close()
            raise
        logger.debug("Got IP discovery packet!")

        # IP is encoded in ASCII, from the forth byte to the first \x00 byte.
        # Find the index of the null byte, starting from 4th.
        ip_start = 4
        ip_end = packet_data.index(0, ip_start)

        our_ip = packet_data[ip_start:ip_end].decode('ascii')
        # Also, we need our local port.
        # Unpack the little-endian (thanks discord!) port from the very end of the voice packet.
        our_port = struct.unpack_from('<H', packet_data, len(packet_data) - 2)[0]

        # Ask the voice websocket to send our SESSION DESCRIPTION packet.
        self.vs_ws.send_select_protocol(our_ip, our_port)
        await self.vs_ws._got_secret_key.wait()
        logger.info("Established connection to Discord's voice servers.")

        self._sock.setblocking(False)

        # We are DONW!

    @classmethod
    async def create(cls, main_client: 'client.Client',
                     gateway, channel: 'dt_channel.Channel') -> 'VoiceClient':
        """
        Creates a new VoiceClient from a channel and a gateway instance.

        :param gateway: The gateway instance to use.
        :param channel: The channel to connect to.
        """
        vs_ws = await VoiceGateway.from_gateway(gw=gateway, channel_id=channel.id, guild_id=channel.guild.id)
        obb = cls(main_client, channel=channel)
        obb.vs_ws = vs_ws
        obb.main_task = await curio.spawn(obb.poll())
        return obb



"""
The HTTP module for Curious.

This code is 100% portable - it will work anywhere that h11 and multidict are also installed.
"""


import logging
import mimetypes
import random
import string
from functools import partial
try:
    import ujson as py_json
except ImportError:
    import json as py_json
import cgi

import multidict
import yarl
import h11
from curio import io
from h11._events import _EventBundle
from multidict import MultiDict
import curio

__version__ = "0.1.0-curious"

logger = logging.getLogger(__name__)


class HTTPError(Exception):
    def __init__(self, *args, **kwargs):
        self.response = kwargs.pop('response')
        super().__init__(*args, **kwargs)


_BOUNDARY_CHARS = string.digits + string.ascii_letters


def encode_multipart(fields, files, boundary=None):
    r"""Encode dict of form fields and dict of files as multipart/form-data.
    Return tuple of (body_string, headers_dict). Each value in files is a dict
    with required keys 'filename' and 'content', and optional 'mimetype' (if
    not specified, tries to guess mime type or uses 'application/octet-stream').

    >>> body, headers = encode_multipart({'FIELD': 'VALUE'},
    ...                                  {'FILE': {'filename': 'F.TXT', 'content': 'CONTENT'}},
    ...                                  boundary='BOUNDARY')
    >>> print('\n'.join(repr(l) for l in body.split('\r\n')))
    '--BOUNDARY'
    'Content-Disposition: form-data; name="FIELD"'
    ''
    'VALUE'
    '--BOUNDARY'
    'Content-Disposition: form-data; name="FILE"; filename="F.TXT"'
    'Content-Type: text/plain'
    ''
    'CONTENT'
    '--BOUNDARY--'
    ''
    >>> print(sorted(headers.items()))
    [('Content-Length', '193'), ('Content-Type', 'multipart/form-data; boundary=BOUNDARY')]
    >>> len(body)
    193

    Copied from: https://code.activestate.com/recipes/578668-encode-multipart-form-data-for-uploading-files-via/
    """

    def escape_quote(s):
        return s.replace(b'"', b'\\"')

    if boundary is None:
        boundary = b''.join(random.choice(_BOUNDARY_CHARS).encode() for i in range(30))
    lines = []

    for name, value in fields.items():
        if isinstance(name, str):
            name = name.encode()
        else:
            name = str(name).encode()

        if isinstance(value, str):
            value = value.encode()
        else:
            value = str(value).encode()

        lines.extend((
            b'--%s' % boundary,
            b'Content-Disposition: form-data; name="%s"' % escape_quote(name),
            b'',
            value,
        ))

    for name, value in files.items():
        filename = value['filename']
        if 'mimetype' in value:
            mimetype = value['mimetype']
        else:
            mimetype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
        lines.extend((
            b'--%s' % boundary,
            b'Content-Disposition: form-data; name="%s"; filename="%s"' % (
                escape_quote(name.encode()), escape_quote(filename.encode())),
            b'Content-Type: %s' % mimetype.encode(),
            b'',
            value['content'],
        ))

    lines.extend((
        b'--%s--' % boundary,
        b'',
    ))
    body = b'\r\n'.join(lines)

    headers = {
        'Content-Type': 'multipart/form-data; boundary=%s' % boundary.decode(),
        'Content-Length': str(len(body)),
    }

    return body, headers


def get_encoding_from_headers(headers: MultiDict):
    """
    Attempts to get the encoding of the request based on the headers.
    """
    content_type = headers.get('content-type')

    if not content_type:
        return None

    content_type, params = cgi.parse_header(content_type)

    if 'charset' in params:
        return params['charset'].strip("'\"")

    if 'text' in content_type:
        return 'ISO-8859-1'


class _EventIterator:
    """
    An iterator that allows you to receive events from a connection.
    """

    def __init__(self, event_source):
        self.event_source = event_source

    async def __aiter__(self):
        return self

    async def __anext__(self):
        event = await self.event_source()

        if type(event) is h11.Data:
            return event.data
        elif type(event) is h11.EndOfMessage:
            raise StopAsyncIteration
        else:
            raise ValueError('Unknown h11 event: %r', event)


class Response:
    """
    A wrapper around the Response from the server.
    """

    def __init__(self, raw_response: h11.Response, h11_request: h11.Request, conn: 'HTTPConnection'):
        self.status_code = raw_response.status_code
        self.http_version = raw_response.http_version.decode('utf-8')

        self.headers = multidict.CIMultiDict(
            (key.decode("utf-8"), value.decode("utf-8")) for key, value in raw_response.headers
        )

        self.h11_request = h11_request
        self.conn = conn

        self.history = None

    def __repr__(self):
        return '<Response [%s]>' % (self.status_code)

    @property
    def url(self):
        """The final URL that was requested."""
        return '{host}:{port}{target}'.format(
            host=self.conn.host,
            port=self.conn.port,
            target=self.h11_request.target.decode('utf-8'),
        )

    async def close(self):
        """
        Closes the current request.

        This should always be called.
        """
        await self.conn.close()

    def raise_for_status(self):
        """Raises HTTPError, if one occurred."""
        http_error_msg = ''

        if 400 <= self.status_code < 500:
            http_error_msg = '%s Client Error for url: %s' % (
                self.status_code, self.url)

        elif 500 <= self.status_code < 600:
            http_error_msg = '%s Server Error for url: %s' % (
                self.status_code, self.url)

        if http_error_msg:
            raise HTTPError(http_error_msg, response=self)

    @property
    def is_redirect(self):
        """Whether the response is a well-formed redirect."""
        return 'location' in self.headers and 301 <= self.status_code < 400

    def iter_chunked(self, maxsize=None):
        """Stream raw response body, maxsize bytes at a time."""
        return _EventIterator(partial(self.conn._next_event, maxsize))

    async def binary(self) -> bytes:
        """Return the full response body as a bytearray."""
        data = b""
        async for chunk in self.iter_chunked():
            data += chunk

        return data

    async def text(self):
        """Return the full response body as a string."""
        data = await self.binary()
        if data is None:
            return None

        # encoding = get_encoding_from_headers(self.headers)

        return data.decode("utf-8")

    async def json(self):
        """Return the full response body as parsed JSON."""
        data = await self.binary()
        if data is None:
            return None

        return py_json.loads(data.decode('utf-8'))


class HTTPConnection:
    """Maries an async socket with an HTTP handler."""

    def __init__(self, host: str, port: int, ssl):
        """
        :param host: The host to connect to.
        :param port: The port to connect to.
        :param ssl: Any SSL context to use.
        """
        self.host = host
        self.port = port
        self.ssl = ssl

        #: The current socket connection.
        self.sock = None  # type: io.Socket
        #: The H11 connection state.
        self.state = None  # type: h11.Connection

    def __repr__(self):
        return '%s(host=%r, port=%r)' % (
            self.__class__.__name__, self.host, self.port)

    async def open(self):
        """
        Opens a connection to the server.
        """
        sock_args = dict(
            host=self.host,
            port=self.port,
        )

        if self.ssl:
            sock_args.update(dict(
                ssl=self.ssl,
                server_hostname=self.host
            ))

        self.sock = await curio.open_connection(**sock_args)
        self.state = h11.Connection(our_role=h11.CLIENT)

        logger.debug('Opened %r', self)

    async def close(self):
        """
        Closes a connection to the server.
        """
        await self.sock.close()

        self.sock = None
        self.state = None

        logger.debug('Closed %r', self)

    async def _send(self, event: _EventBundle):
        """
        Sends an event to the server.

        :param event: The event to send.
        """
        # logger.debug("Sending event: %s", event)

        data = self.state.send(event)
        await self.sock.sendall(data)

    async def _next_event(self, maxsize: int=None) -> _EventBundle:
        """
        Gets the next event from the connection.

        This will automatically read data off of the connection.

        :param maxsize: The maximum size to read off the socket.
            If not specified, 2048 will be used.
        :return:
        """

        if not maxsize:
            maxsize = 2048

        while True:
            event = self.state.next_event()

            # logger.debug("Received event: %s", event)

            if event is h11.NEED_DATA:
                data = await self.sock.recv(maxsize)
                self.state.receive_data(data)
                continue

            return event

    async def request(self, h11_request: h11.Request, data=None) -> h11.Response:
        """
        Makes a request.

        :param h11_request: The :class:`h11.Request` object to use.
        :param data: Any request data to send to the server.
        """
        # First, send the base request.
        await self._send(h11_request)

        # Send data chunks.
        if data is not None:
            await self._send(h11.Data(data=data))

        # Finally, send the EOF.
        await self._send(h11.EndOfMessage())

        # Read the response off of the server now.
        event = await self._next_event()

        assert isinstance(event, h11.Response)

        return event


def _prepare_request(method: str, url: yarl.URL, *,
                     params=None, headers=None,
                     body=None,
                     json: dict = None, files: dict = None):
    """
    Prepares a new request, creating a :class:`h11.Request` object to be sent.
    Additionally, this also serializes the body into a bytestring ready to be sent.

    :param method: The method of the request.
    :param url: The URL to request.
    :param headers: Optional. The headers of the request.
        This should be a dict or a MultiDict.
    :param body: Optional. Any body data to send.
        If this is a dict, it is encoded as multipart/form-data.
    :param json: Optional. A dict that should be JSON encoded and sent as the body.
    """
    url = yarl.URL(url)

    if params:
        query_vars = list(url.query.items()) + list(params.items())
        url = url.with_query(query_vars)

    target = str(url.relative())

    if headers is None:
        headers = MultiDict()
    elif not isinstance(headers, MultiDict):
        headers = MultiDict(headers)

    headers.setdefault('Host', url.raw_host)

    if json is not None:
        body = py_json.dumps(json).encode()
    else:
        # Check if the body is a dict.
        # If so, send it as `multipart/form-data`.
        if isinstance(body, dict):
            body, _h = encode_multipart(body, files)
            headers.update(_h)

    if body is not None:
        if "Content-Length" not in headers:
            headers["Content-Length"] = str(len(body)).encode('utf-8')

    if json is not None:
        headers["Content-Type"] = b"application/json"

    if "Content-Length" not in headers:
        headers["Content-Length"] = b"0"

    if "User-Agent" not in headers:
        headers["User-Agent"] = "curio_http/{} curio/{}".format(__version__, curio.__version__)

    if not all(isinstance(header, (str, bytes)) for header in headers.keys()):
        raise ValueError("Header keys must be str or bytes")

    if not all(isinstance(header, (str, bytes)) for header in headers.values()):
        raise ValueError("Header values must be str or bytes")

    h11_request = h11.Request(
        method=method,
        target=target,
        headers=list(headers.items())
    )

    return h11_request, body


class ClientSession:
    """
    The session class that is used to send data to the server.
    """
    def __init__(self):
        #: A dictionary of all open connections.
        self.open_connections = []

        #: Any headers that are to be sent on every request.
        self.headers = MultiDict()
        self.headers.setdefault("User-Agent", "curio-http/{} (bundled with curious)".format(__version__))

    async def __aenter__(self) -> 'ClientSession':
        return self

    async def __aexit__(self, exc_type, exc, tb):
        for conn in self.open_connections:
            await conn.close()

    async def _request(self, method, url, *args, **kwargs):
        """
        Makes a request to the server.

        This method is internal - use `request` instead.
        """
        headers = MultiDict(self.headers.copy())
        headers.extend(kwargs.get("headers", {}))
        kwargs["headers"] = headers

        h11_request, body = _prepare_request(
            method, url, *args, **kwargs)

        conn = HTTPConnection(
            host=url.raw_host,
            port=url.port,
            ssl=url.scheme == 'https',
        )

        await conn.open()

        self.open_connections.append(conn)

        raw_response = await conn.request(h11_request, body)

        return Response(raw_response, h11_request, conn)

    async def request(self, method: str, url: str, *args,
                      allow_redirects=False, **kwargs):
        """
        Performs a HTTP request.

        :param method: The HTTP method of the request.
        :param url: The URL of the request.
        :param allow_redirects: Should redirects be automatically followed?
        """
        url = yarl.URL(url)

        response = await self._request(method, url, *args, **kwargs)

        if allow_redirects:
            history = []

            while response.is_redirect:
                history.append(response)

                # Redirects can be relative.
                new_url = url.join(yarl.URL(response.headers['location']))

                response = await self._request(method, new_url)

            response.history = history

        return response

    def get(self, *args, allow_redirects=True, **kwargs):
        """Perform HTTP GET request."""
        return self.request(
            'GET', *args, allow_redirects=allow_redirects, **kwargs)

    def post(self, *args, **kwargs):
        """Perform HTTP POST request."""
        return self.request('POST', *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.request("PUT", *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self.request("PATCH", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.request("DELETE", *args, **kwargs)



"""
The core plugin.

This contains a help command that can be used.
"""
from curious.dataclasses.embed import Embed

from curious.commands import command, plugin
from curious.commands.context import Context


class _Core(plugin.Plugin):
    name = "Core"
    _include_in_scan = False

    # region embeds
    async def _help_all_with_embeds(self, ctx: Context):
        # display help for all commands
        em = Embed(title=ctx.bot.description)
        if ctx.prefix is not None:
            em.description = "Type {}help <command> for more information.".format(ctx.prefix)
        else:
            em.description = "Use `help <command>` for more information."

        for plugin in ctx.bot.plugins.copy().values():
            gen = list(ctx.bot.get_commands_for(plugin))
            cmds = sorted(gen, key=lambda c: c.name)

            names = []
            for cmd in cmds:
                failed, result = await cmd.can_run(ctx)
                if not failed:
                    names.append("`{}`".format(cmd.name))

            names = ", ".join(names)
            if not names:
                continue

            em.add_field(name=plugin.name, value=names, inline=False)

        return em

    async def _help_for_command_with_embeds(self, ctx: Context, command_for_help: str):
        initial_name = command_for_help
        parts = command_for_help.split(" ")
        command = parts[0]

        # get the initial command object
        command_obb = ctx.bot.get_command(command)
        if command_obb is not None:
            for token in parts[1:]:
                command = token
                command_obb = command_obb.find_subcommand(token)
                if command_obb is None:
                    # exit early
                    break

        if not command_obb:
            em = Embed(title=command, description="Command not found.", colour=0xe74c3c)
        else:
            # Check if this was an alias.
            if command_obb.name != command:
                title = "{} (alias for `{}`)".format(command, command_obb.name)
            else:
                title = command_obb.name
            em = Embed(title=title)
            em.description = command_obb.get_help(ctx, command)  # Stop. Get help.

            # If this has other aliases, add them.
            if len(command_obb.aliases) != 1:
                b = "\n".join("`{}`".format(n) for n in command_obb.aliases if n != command)
                em.add_field(name="Aliases", value=b)

            # Check if they can run the command.
            if len(command_obb.invokation_checks) > 0:
                failed, result = await command_obb.can_run(ctx)
                if failed:
                    em.description += "\n\nYou **cannot** run this command. (`{}` checks failed.)".format(failed)
                    em.colour = 0xFF0000

            usage = command_obb.get_usage(ctx, initial_name)
            em.add_field(name="Usage", value=usage)

            sbb = "\n".join([subcommand.name for subcommand in command_obb.subcommands])
            if sbb:
                em.add_field(name="Subcommands", value=sbb)

        return em

    async def _help_with_embeds(self, ctx: Context, command_for_help: str):
        # help command WITH embeds
        if command_for_help is None:
            embed = await self._help_all_with_embeds(ctx)
        else:
            embed = await self._help_for_command_with_embeds(ctx, command_for_help)

        await ctx.channel.send(embed=embed)

    # endregion

    async def _help_without_embeds(self, ctx: Context, command: str = None):
        """
        The default help command without embeds.
        """
        if not command:
            base = "**Commands:**\nUse `{}help <command>` for more information about each command.\n\n".format(
                ctx.prefix)

            for num, plugin in enumerate(ctx.bot.plugins.copy().values()):
                gen = list(ctx.bot.get_commands_for(plugin))
                cmds = sorted(gen, key=lambda c: c.name)

                names = " **|** ".join("`{}`".format(cmd.name) for cmd in cmds)
                base += "**{}. {}**: {}\n".format(num + 1, plugin.name, names)

            msg = base
        else:
            initial_name = command
            parts = command.split(" ")
            command = parts[0]

            # get the initial command object
            command_obb = ctx.bot.get_command(command)
            if command_obb:
                for token in parts[1:]:
                    command = token
                    command_obb = command_obb.find_subcommand(token)
                    if command_obb is None:
                        # exit early
                        break

            if command_obb is None:
                msg = "Command not found."
            else:
                base = command_obb.get_usage(ctx, initial_name) + "\n\n"
                base += "{}".format(command_obb.get_help(ctx, initial_name))
                msg = "```{}```".format(base)

        msg += "\n**For a better help command, give the bot Embed Links permission.**"

        await ctx.message.channel.send(msg)

    @command(name="help", overridable=True)
    async def _help_command(self, ctx: Context, *, command_for_help: str=None):
        """
        Displays help for a command.
        """
        if ctx.channel.permissions(ctx.guild.me).embed_links:
            await self._help_with_embeds(ctx, command_for_help)
        else:
            await self._help_without_embeds(ctx, command_for_help)




"""
Voice connection utilities.
"""
try:
    import opuslib
except ImportError as e:
    raise ImportError("You must install `opuslib` to use curious' voice") from e

try:
    import nacl
except ImportError as e:
    raise ImportError("You must install `pynacl` to use curious' voice") from e

from .voice_player import VoicePlayer
from .voice_client import VoiceClient



import inspect
import typing

from curious.commands.context import Context
from curious.commands.exc import CheckFailureError, MissingArgumentError, CommandInvokeError, ConversionFailedError

# Default converters.
from curious.dataclasses import channel as dt_channel
from curious.dataclasses import member as dt_member
from curious.util import replace_quotes


def convert_member(ctx: Context, arg: str):
    if arg.startswith("<@") and arg.endswith(">"):
        # Parse the mention out
        id = arg[2:-1]
        if id[0] == "!":  # nicknames
            id = id[1:]

        try:
            id = int(id)
        except ValueError:
            raise ConversionFailedError(ctx, arg, dt_member.Member)

        member = ctx.guild.members.get(id)
        if not member:
            # todo: better error
            raise ConversionFailedError(ctx, arg, dt_member.Member)
    else:
        member = ctx.guild.find_member(arg)
        if not member:
            raise ConversionFailedError(ctx, arg, dt_member.Member)

    return member


def convert_channel(ctx: Context, arg: str):
    if arg.startswith("<#") and arg.endswith(">"):
        id = arg[2:-1]

        try:
            id = int(id)
        except ValueError:
            raise ConversionFailedError(ctx, arg, dt_channel.Channel)

        channel = ctx.guild.channels.get(id)
        if not channel:
            raise ConversionFailedError(ctx, arg, dt_channel.Channel)
    else:
        try:
            channel = next(filter(lambda c: c.name == arg, ctx.guild.channels), None)
            if channel is None:
                channel = next(filter(lambda c: c.id == int(arg), ctx.guild.channels))
        except (StopIteration, ValueError):
            raise ConversionFailedError(ctx, arg, dt_channel.Channel)

    return channel


def convert_int(ctx: Context, arg: str):
    try:
        return int(arg)
    except ValueError as e:
        raise ConversionFailedError(ctx, arg, int)


class Command(object):
    """
    A command object represents a command.
    """
    converters = {
        int: convert_int,
    }

    def __init__(self, cbl, *,
                 name: str = None, aliases: typing.List[str] = None,
                 invokation_checks: list = None, group: bool = False,
                 overridable: bool=False):
        """
        :param cbl: The callable to use.
        :param name: The name of this command.
            If this isn't provided, it will be automatically determined from the callable name.

        :param aliases: A list of aliases that this command can be called as.
        :param group: Is this the root command for a group?
        :param overridable: Can this command be overridden?
        """
        self.callable = cbl

        self.name = name
        if not self.name:
            # Use the __name__ of the callable instead
            # This isn't always accurate, but you should provide `name=` if you want that.
            self.name = self.callable.__name__

        self.aliases = [self.name] + (aliases if aliases else [])

        # Pre-calculate the function signature.
        # Saves a few MS every command in
        self._signature = inspect.signature(self.callable)

        #: The plugin instance to pass to this command.
        #: If this is None, it is assumed the command invoker does not need an instance.
        self._instance = None

        #: A list of invokation checks that must all return True before the underlying function is run.
        self.invokation_checks = invokation_checks if invokation_checks else []

        #: Is this a group command?
        self._is_group = group

        #: Is this command overridable?
        self._overridable = overridable

        #: The subcommands for this command.
        self.subcommands = []  # type: typing.List[Command]

        if self._is_group:
            # Scan our callable for subcommands.
            factories = self.callable.subcommands
            for f in factories:
                subcommand = f()  # type: Command
                self.subcommands.append(subcommand)

        if dt_channel.Channel not in self.converters:
            self.converters[dt_channel.Channel] = convert_channel

        if dt_member.Member not in self.converters:
            self.converters[dt_member.Member] = convert_member

    @property
    def instance(self):
        return self._instance

    @instance.setter
    def instance(self, value):
        if value is None:
            self._instance = None
            return

        # add the plugin_check func as a local check
        self.invokation_checks.append(value.plugin_check)
        self._instance = value

        for subcommand in self.subcommands:
            subcommand.instance = value

    def __repr__(self):
        return "<Command name='{}' plugin='{}' subcommands={}>".format(self.name, self.instance, self.subcommands)

    @classmethod
    def add_converter(cls, type_: typing.Type, func: typing.Callable[[Context, typing.Any], str]):
        """
        Adds a converter to the local dict of converters.

        :param func: The function to add.
        :param type_: The type to convert.
        """
        cls.converters[type_] = func

    def find_subcommand(self, name: str) -> 'typing.Union[None, Command]':
        """
        Finds a subcommand by name.

        :param name: The name of the command to fetch.
        :return: A command object, if found.
        """
        for command in self.subcommands:
            if command.can_be_invoked_by(name):
                return command

    def get_usage(self, ctx: Context, invoked_as: str) -> str:
        """
        :return: The usage text for this command.
        """
        if ctx.prefix is not None:
            final = ["{}{}".format(ctx.prefix, invoked_as)]
        else:
            final = [invoked_as]

        for n, (name, param) in enumerate(self._signature.parameters.items()):
            # always skip the first arg
            if n == 0:
                continue

            if self.instance is not None:
                # skip the 2nd arg if we're bound to a plugin
                if n == 1:
                    continue

            # switch based on kind (again...)
            assert isinstance(param, inspect.Parameter)
            if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.POSITIONAL_ONLY]:
                if param.annotation is not inspect.Parameter.empty:
                    s = "<{}: {}>".format(param.name, param.annotation.__name__)
                else:
                    s = "<{}>".format(param.name)

            elif param.kind in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_POSITIONAL]:
                if param.default is not inspect.Parameter.empty:
                    if param.annotation is not inspect.Parameter.empty:
                        s = "[{}: {} (default: {})]".format(param.name, param.annotation.__name__,
                                                            repr(param.default))
                    else:
                        s = "[{} (default: {})]".format(param.name, repr(param.default))
                else:
                    if param.annotation is not inspect.Parameter.empty:
                        s = "<{}: {}>".format(param.name, param.annotation.__name__)
                    else:
                        s = "<{}>".format(param.name)
            else:
                s = ""

            final.append(s)

        return " ".join(final)

    def get_help(self, ctx: Context, invoked_as: str) -> str:
        """
        :return: The help text for this command.
        """
        doc = inspect.getdoc(self.callable)
        if not doc:
            return "This command has no help."
        else:
            return doc

    def _lookup_converter(self, thing, type_: type) -> typing.Callable[[Context, object], str]:
        """
        Gets a converter for the specified type.
        This is provided as a function for command subclasses to be able to override.
        """
        if thing is None:
            return lambda c, i: i
        return self.converters.get(type_, lambda ctx, x: str(x))

    async def _convert(self, ctx, *args):
        """
        Converts the arguments passed from discord into the function, based on it's type signature.
        """
        final_args = []
        final_kwargs = {}

        args_it = iter(args)

        for n, (name, param) in enumerate(self._signature.parameters.items()):
            if n == 0:
                # Don't convert the `self/ctx` argument.
                continue

            if self.instance is not None:
                if n == 1:
                    # Don't convert the bound `ctx` argument.
                    continue

            assert isinstance(param, inspect.Parameter)
            # We loop over the signature parameters because it's easier to use those to consume.
            # Get the next argument from args.
            try:
                arg = next(args_it)
            except StopIteration as e:
                # not good!
                # If we're a *arg format, we can safely handle this, or if we have a default.
                if param.kind == inspect.Parameter.VAR_POSITIONAL:
                    break

                if param.default is inspect.Parameter.empty:
                    raise MissingArgumentError(ctx, param.name) from e

                arg = param.default

            # Begin the consumption!
            if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.POSITIONAL_ONLY]:
                # Only add it to the final_args, then continue the loop.
                arg = replace_quotes(arg)
                converter = self._lookup_converter(arg, param.annotation)
                final_args.append(converter(ctx, arg))
                continue

            if param.kind in [inspect.Parameter.KEYWORD_ONLY]:
                # Only add it to final_kwargs.
                # This is a consume all operation, so we eat all of the arguments.
                f = [arg]

                while True:
                    try:
                        next_arg = next(args_it)
                    except StopIteration:
                        break

                    f.append(next_arg)

                converter = self._lookup_converter(arg, param.annotation)
                if len(f) == 1:
                    final_kwargs[param.name] = converter(ctx, f[0])
                else:
                    final_kwargs[param.name] = converter(ctx, " ".join(f))
                continue

            if param.kind in [inspect.Parameter.VAR_POSITIONAL]:
                # This *shouldn't* be called on `*` arguments, but we can't be sure.
                # Special case - consume ALL the arguments.
                f = [arg]

                while True:
                    try:
                        next_arg = next(args_it)
                    except StopIteration:
                        break

                    f.append(next_arg)

                converter = self._lookup_converter(arg, param.annotation)
                final_args.append(converter(ctx, " ".join(f)))

                # We *could* get more arguments.
                # I mean, there's no more argument data to consume.
                # So if you haven't set a default, it won't show anything.
                # Continue anyway, though.
                continue

            if param.kind in [inspect.Parameter.VAR_KEYWORD]:
                # no
                continue

        return final_args, final_kwargs

    async def can_run(self, ctx: Context) -> typing.Tuple[int, typing.List[typing.Tuple[str, typing.Any]]]:
        """
        Checks if this command can run.

        This will return a two-item tuple: the number of checks that failed, the results of the checks.

        :param ctx: The command context.
        :return: The tuple described above.
        """
        failed, results = len(self.invokation_checks), []

        for check in self.invokation_checks:
            try:
                result = check(ctx)
                if inspect.isawaitable(result):
                    result = await result
            except CheckFailureError as e:
                result = repr(e)

            if not isinstance(result, CheckFailureError) and result:  # truthy value means succeeded
                failed -= 1

            results.append((check.__name__, result))

        return failed, results

    def can_be_invoked_by(self, token: str) -> bool:
        """
        Checks if this command can be invoked by the specified token.

        :param token: The token to check.
        :return: True if it can, False if it can't.
        """
        if self.name == token:
            return True

        if token in self.aliases:
            return True

        return False

    async def invoke(self, ctx: Context, *args):
        failed, _ = await self.can_run(ctx)
        if failed:
            raise CheckFailureError(ctx, _)

        # check if we can run a subcommand, first.
        try:
            subcommand_token = args[0]

            for subcommand in self.subcommands:
                if subcommand.can_be_invoked_by(subcommand_token):
                    await subcommand.invoke(ctx, *args[1:])
                    # never invoke groups if a subcommand hit
                    return

        except IndexError:
            # no subcommand to invoke
            pass

        final_args, final_kwargs = await self._convert(ctx, *args)

        try:
            if self.instance is not None:
                await self.callable(self.instance, ctx, *final_args, **final_kwargs)
            else:
                await self.callable(ctx, *final_args, **final_kwargs)
        except Exception as e:
            raise CommandInvokeError(ctx) from e



from curious.dataclasses import user as dt_user
from curious.dataclasses import guild as dt_guild
from curious.dataclasses import channel as dt_channel


class VoiceState(object):
    """
    Represents the voice state of a user.
    """
    __slots__ = ("_user_id", "user", "_guild_id", "guild", "_channel_id", "channel", "_self_mute", "_server_mute",
                 "_self_deaf", "_server_deaf")

    def __init__(self, user: 'dt_user.User', **kwargs):
        self._user_id = int(kwargs.pop("user_id", 0)) or None
        #: The user object this voice state is for.
        self.user = user

        self._guild_id = int(kwargs.pop("guild_id", 0)) or None

        #: The guild this voice state is associated with.
        self.guild = None  # type: dt_guild.Guild

        self._channel_id = int(kwargs.pop("channel_id", 0)) or None
        #: The voice channel this member is in.
        self.channel = None  # type: dt_channel.Channel

        # Internal state values.
        self._self_mute = kwargs.pop("self_mute", False)
        self._server_mute = kwargs.pop("mute", False)
        self._self_deaf = kwargs.pop("self_deaf", False)
        self._server_deaf = kwargs.pop("deaf", False)

    @property
    def muted(self) -> bool:
        """
        :return: If this user is muted or not.
        """
        return self._server_mute or self._self_mute

    @property
    def deafened(self) -> bool:
        """
        :return: If this user is deafened or not.
        """
        return self._server_deaf or self._self_deaf

    @property
    def member(self):
        """
        :return: The member this is associated with.
        """
        return self.guild.members[int(self._user_id)]

    def __repr__(self):
        return "<VoiceState user={} deaf={} mute={} channel={}>".format(self.user, self.deafened, self.muted,
                                                                        self.channel)

    async def mute(self):
        """
        Server mutes this member on the guild.
        """
        await self.guild.change_voice_state(self.member, mute=True)

    async def unmute(self):
        """
        Server unmutes this member on the guild.
        """
        await self.guild.change_voice_state(self.member, mute=False)

    async def deafen(self):
        """
        Server deafens this member on the guild.
        """
        await self.guild.change_voice_state(self.member, deaf=True)

    async def undeafen(self):
        """
        Server undeafens this member on the guild.
        """
        await self.guild.change_voice_state(self.member, deaf=False)



from math import ceil

import curio
import typing
from types import MappingProxyType

from curious.dataclasses import channel, emoji as dt_emoji, invite as dt_invite, member as dt_member, \
    permissions as dt_permissions, role, user as dt_user, voice_state as dt_vs, webhook as dt_webhook
from curious.dataclasses.bases import Dataclass
from curious.dataclasses.status import Game
from curious.exc import CuriousError, HierachyError, PermissionsError
from curious.util import AsyncIteratorWrapper, base64ify

try:
    from curious.voice import voice_client
except ImportError:
    VoiceClient = None
else:
    VoiceClient = voice_client.VoiceClient


class Guild(Dataclass):
    """
    Represents a guild object on Discord.

    :ivar id: The ID of this guild.
    """

    __slots__ = ("id", "unavailable", "name", "_icon_hash", "_splash_hash", "_owner_id", "_afk_channel_id",
                 "afk_timeout", "region", "mfa_level", "verification_level", "shard_id", "_roles", "_members",
                 "_channels", "_emojis", "_finished_chunking", "member_count", "large", "_chunks_left", "voice_client",
                 )

    def __init__(self, bot, **kwargs):
        """
        Creates a new Guild object.
        """
        super().__init__(kwargs.pop("id"), bot)

        #: If the guild is unavailable or not.
        #: If this is True, many fields return `None`.
        self.unavailable = kwargs.get("unavailable", False)

        # Placeholder values.
        #: The name of this guild.
        self.name = None  # type: str

        #: The icon hash of this guild.
        #: Used to construct the icon URL later.
        self._icon_hash = None  # type: str

        #: The splash hash of this guild.
        #: Used to construct the splash URL later.
        self._splash_hash = None  # type: str

        #: The owner ID of this guild.
        self._owner_id = None  # type: int

        #: The AFK channel ID of this guild.
        self._afk_channel_id = None  # type: int

        #: The AFK timeout for this guild.
        self.afk_timeout = None  # type: int

        #: The voice region of this guild.
        self.region = None  # type: str

        #: The MFA level of this guild.
        self.mfa_level = 0  # type: int

        #: The verification level of this guild.
        self.verification_level = 0  # type: int

        #: The shard ID this guild is associated with.
        self.shard_id = None

        #: The roles that this guild has.
        self._roles = {}

        #: The members of this guild.
        self._members = {}

        #: The channels of this guild.
        self._channels = {}

        #: The emojis that this guild has.
        self._emojis = {}

        #: The number of numbers this guild has.
        #: This is automatically updated.
        self.member_count = 0  # type: int

        #: Is this guild a large guild?
        self.large = False  # type: bool

        #: Has this guild finished chunking?
        self._finished_chunking = curio.Event()
        self._chunks_left = 0

        #: The current voice client associated with this guild.
        self.voice_client = None  # type: VoiceClient

        self.from_guild_create(**kwargs)

    def _copy(self):
        obb = object.__new__(self.__class__)

        obb.unavailable = self.unavailable
        obb.name = self.name
        obb._icon_hash = self._icon_hash
        obb._splash_hash = self._splash_hash
        obb._owner_id = self._owner_id
        obb.region = self.region
        obb.shard_id = self.shard_id
        obb._roles = self._roles.copy()
        obb._members = self._members.copy()
        obb._channels = self._members.copy()
        obb.member_count = self.member_count
        obb.large = self.large
        obb._afk_channel_id = self._afk_channel_id
        obb.afk_timeout = self.afk_timeout
        obb.mfa_level = self.mfa_level
        obb._emojis = self._emojis.copy()

        return obb

    @property
    def channels(self) -> 'typing.Mapping[channel.Channel]':
        """
        :return: A list of :class:`curious.dataclasses.channel.Channel` that represent the channels on this guild.
        """
        return MappingProxyType(self._channels)

    @property
    def members(self) -> 'typing.Mapping[int, dt_member.Member]':
        """
        :return: A mapping of :class:`Member` that represent members on this guild.
        """
        return MappingProxyType(self._members)

    @property
    def roles(self) -> 'typing.Mapping[int, dt_member.Member]':
        """
        :return: A mapping of :class:`Role` on this guild.
        """
        return MappingProxyType(self._roles)

    @property
    def emojis(self) -> 'typing.Mapping[int, dt_emoji.Emoji]':
        """
        :return: A mapping of :class:`Emoji` on this guild.
        """
        return MappingProxyType(self._emojis)

    @property
    def owner(self) -> 'dt_member.Member':
        """
        :return: A :class:`curious.dataclasses.member.Member` object that represents the owner of this guild.
        """
        return self._members[self._owner_id]

    @property
    def me(self) -> 'dt_member.Member':
        """
        :return: A :class:`curious.dataclasses.member.Member` object that represents the current user in this guild.
        """
        return self._members[self._bot.user.id]

    @property
    def default_channel(self) -> 'channel.Channel':
        """
        :return: A :class:`curious.dataclasses.channel.Channel` that represents the default channel of this guild.
        """
        return self._channels[self.id]

    @property
    def default_role(self) -> 'role.Role':
        """
        :return: A :class:`curious.dataclasses.role.Role` that represents the default role of this guild.
        """
        return self._roles[self.id]

    @property
    def afk_channel(self) -> 'channel.Channel':
        """
        :return: A :class:`Channel` representing the AFK channel for this guild.
        """
        try:
            return self._channels[self._afk_channel_id]
        except IndexError:
            # the afk channel CAN be None
            return None

    @property
    def embed_url(self) -> str:
        """
        Gets the default embed url for this guild.
        The widget must be enabled.
        
        :return: The embed URL for this guild. 
        """
        return (self._bot.http.GUILD_BASE + "/embed.png").format(guild_id=self.id)

    def get_embed_url(self, *, style: str = "banner1") -> str:
        """
        Gets an embed URL for this guild in a specified style.
        
        :param style: The style to get. 
        :return: The embed URL for this guild.
        """
        valid_styles = ('banner1', 'banner3', 'banner2', 'shield', 'banner4')
        if style not in valid_styles:
            raise ValueError("Style must be in {}".format(valid_styles))

        return self.embed_url + "?style={}".format(style)

    def find_member(self, search_str: str) -> 'dt_member.Member':
        """
        Attempts to find a member in this guild by name#discrim.
        This will also search nicknames.

        The discrim is optional, but if provided allows better matching.

        :param search_str: The name#discrim pair to search for.
        :return: A :class:`Member` object that represents the member, or None if no member could be found.
        """
        sp = search_str.rsplit("#", 1)
        if len(sp) == 1:
            # Member name only :(
            predicate = lambda member: member.user.name == sp[0] or member.nickname == sp[0]
        else:
            # Discriminator too!
            # Don't check nicknames for this.
            predicate = lambda member: member.user.name == sp[0] and member.user.discriminator == sp[1]

        filtered = filter(predicate, self.members.values())
        try:
            return next(filtered)
        except StopIteration:
            return None

    def start_chunking(self):
        self._finished_chunking.clear()
        self._chunks_left = ceil(self.member_count / 1000)

    async def wait_until_chunked(self):
        """
        Waits until the guild has finished chunking.

        Useful for when you join a big guild.
        """
        await self._finished_chunking.wait()

    def _handle_member_chunk(self, members: list):
        """
        Handles a chunk of members.
        """
        if self._chunks_left >= 1:
            # We have a new chunk, so decrement the number left.
            self._chunks_left -= 1

        for member_data in members:
            id = int(member_data["user"]["id"])
            if id in self._members:
                member_obj = self._members[id]
            else:
                member_obj = dt_member.Member(self._bot, **member_data)
            for role_ in member_data.get("roles", []):
                role_obj = self._roles.get(int(role_))
                if role_obj:
                    member_obj._roles[role_obj.id] = role_obj

            member_obj.guild = self
            self._members[member_obj.id] = member_obj

    def _handle_emojis(self, emojis: list):
        """
        Handles the emojis for this guild.
        """
        for emoji in emojis:
            emoji_obj = dt_emoji.Emoji(**emoji)
            emoji_obj.guild = self
            for role_id in emoji_obj._role_ids:
                emoji_obj.roles.append(self._roles.get(int(role_id)))

            self._emojis[emoji_obj.id] = emoji_obj

    def from_guild_create(self, **data: dict) -> 'Guild':
        """
        Populates the fields from a GUILD_CREATE event.

        :param data: The GUILD_CREATE data to use.
        """
        self.unavailable = data.pop("unavailable", False)

        if self.unavailable:
            # We can't use any of the extra data here, so don't bother.
            return self

        self.name = data.get("name")  # type: str
        self._icon_hash = data.get("icon")  # type: str
        self._splash_hash = data.get("splash")  # type: str
        self._owner_id = int(data.get("owner_id"))  # type: int
        self.large = data.get("large", False)
        self.region = data.get("region")
        afk_channel_id = data.get("afk_channel_id")
        if afk_channel_id:
            afk_channel_id = int(afk_channel_id)
        self._afk_channel_id = afk_channel_id
        self.afk_timeout = data.get("afk_timeout")
        self.verification_level = data.get("verification_level")
        self.mfa_level = data.get("mfa_level")

        self.member_count = data.get("member_count", 0)

        # Create all the Role objects for the server.
        for role_data in data.get("roles", []):
            role_obj = role.Role(self._bot, **role_data)
            role_obj.guild = self
            self._roles[role_obj.id] = role_obj

        # Create all the Member objects for the server.
        self._handle_member_chunk(data.get("members", []))

        for presence in data.get("presences", []):
            member_id = int(presence["user"]["id"])
            member_obj = self._members.get(member_id)

            if not member_obj:
                continue

            game = presence.get("game", {})
            if game is not None:
                member_obj.game = Game(**game)

            member_obj.status = presence.get("status")

        # Create all of the channel objects.
        for channel_data in data.get("channels", []):
            channel_obj = channel.Channel(self._bot, guild=self, **channel_data)
            channel_obj.guild = self
            self._channels[channel_obj.id] = channel_obj

        # Create all of the voice states.
        for vs_data in data.get("voice_states", []):
            user_id = int(vs_data.get("user_id", 0))
            member = self.members.get(user_id)
            if not member:
                # o well
                continue

            voice_state = dt_vs.VoiceState(member.user, **vs_data)

            vs_channel = self._channels.get(int(vs_data.get("channel_id", 0)))
            voice_state.channel = vs_channel
            voice_state.guild = self
            member.voice = voice_state

        # Create all of the emoji objects for the server.
        self._handle_emojis(data.get("emojis", []))

    @property
    def bans(self) -> 'typing.AsyncIterator[dt_user.User]':
        return AsyncIteratorWrapper(self._bot, self.get_bans())

    @property
    def invites(self) -> 'typing.AsyncIterator[dt_invite.Invite]':
        return AsyncIteratorWrapper(self._bot, self.get_invites())

    @property
    def icon_url(self) -> str:
        """
        :return: The icon URL for this guild, or None if one isn't set.
        """
        if self._icon_hash:
            return "https://cdn.discordapp.com/icons/{}/{}.webp".format(self.id, self._icon_hash)

    @property
    def splash_url(self) -> str:
        """
        :return: The splash URL for this guild, or None if one isn't set.
        """
        if self._splash_hash:
            return "https://cdn.discordapp.com/splashes/{}/{}.webp".format(self.id, self._splash_hash)

    # Guild methods.
    async def leave(self):
        """
        Leaves the guild.
        """
        await self._bot.http.leave_guild(self.id)

    async def connect_to_voice(self, channel: 'channel.Channel') -> VoiceClient:
        """
        Connects to a voice channel in this guild.

        :param channel: The channel to connect to.
        :return: The :class:`VoiceClient` that was connected to this guild.
        """
        if VoiceClient is None:
            raise RuntimeError("Cannot to voice - voice support is not installed")

        if channel.guild != self:
            raise CuriousError("Cannot use channel from a different guild")

        if self.voice_client is not None and self.voice_client.open:
            raise CuriousError("Voice client already exists in this guild")

        gw = self._bot._gateways[self.shard_id]
        self.voice_client = await VoiceClient.create(self._bot, gw, channel)
        await self.voice_client.connect()
        return self.voice_client

    async def get_invites(self) -> 'typing.List[dt_invite.Invite]':
        """
        Gets the invites for this guild.
        :return: A list of invite objects.
        """
        invites = await self._bot.http.get_invites_for(self.id)
        invites = [dt_invite.Invite(self._bot, **i) for i in invites]

        return invites

    async def get_bans(self) -> 'typing.List[dt_user.User]':
        """
        Gets the bans for this guild.
        :return: A list of User objects, one for each ban.
        """
        if not self.me.guild_permissions.ban_members:
            raise PermissionsError("ban_members")

        bans = await self._bot.http.get_bans(self.id)
        users = []

        for user_data in bans:
            # TODO: Audit log stuff, if it ever comes out.
            user_data = user_data.get("user", None)
            users.append(dt_user.User(self._bot, **user_data))

        return users

    async def kick(self, victim: 'dt_member.Member'):
        """
        Kicks somebody from the guild.

        :param victim: The member to kick.
        """
        if not self.me.guild_permissions.kick_members:
            raise PermissionsError("kick_members")

        if victim.guild != self:
            raise ValueError("Member must be from this guild (try `member.user` instead)")

        if victim.top_role >= self.me.top_role:
            raise HierachyError("Top role is equal to or lower than victim's top role")

        victim_id = victim.user.id

        await self._bot.http.kick_member(self.id, victim_id)

    async def ban(self, victim: 'typing.Union[dt_member.Member, dt_user.User]', *,
                  delete_message_days: int = 7):
        """
        Bans somebody from the guild.

        This can either ban a Member, in which they must be in the guild. Or this can ban a User, which does not need
        to be in the guild.

        Example for banning a member:

        .. code:: python

            member = guild.members[66237334693085184]
            await guild.ban(member)

        Example for banning a user:

        .. code:: python

            user = await client.get_user(66237334693085184)
            await guild.ban(user)

        :param victim: The person to ban.
        :param delete_message_days: The number of days to delete messages.
        """
        if not self.me.guild_permissions.ban_members:
            raise PermissionsError("ban_members")

        if isinstance(victim, dt_member.Member):
            if self.owner == victim:
                raise HierachyError("Cannot ban the owner")

            if victim.guild != self:
                raise ValueError("Member must be from this guild (try `member.user` instead)")

            if victim.top_role >= self.me.top_role:
                raise HierachyError("Top role is equal to or lower than victim's top role")

            victim_id = victim.user.id

        elif isinstance(victim, dt_user.User):
            victim_id = victim.id

        else:
            raise TypeError("Victim must be a Member or a User")

        await self._bot.http.ban_user(guild_id=self.id, user_id=victim_id, delete_message_days=delete_message_days)

    async def unban(self, user: 'dt_user.User'):
        """
        Unbans a user from this guild.

        Example for unbanning the first banned user:

        .. code:: python

            user = next(await guild.get_bans())
            await guild.unban(user)

        To unban an arbitrary user, use :meth:`Client.get_user`.

        .. code:: python

            user = await client.get_user(66237334693085184)
            await guild.unban(user)

        :param user: The user to forgive and unban.
        """
        if not self.me.guild_permissions.ban_members:
            raise PermissionsError("ban_members")

        forgiven_id = user.id

        await self._bot.http.unban_user(self.id, forgiven_id)

    async def get_webhooks(self) -> 'typing.List[dt_webhook.Webhook]':
        """
        Gets the webhooks for this guild.

        :return: A list of :class:`Webhook` objects for the guild.
        """
        webhooks = await self._bot.http.get_webhooks_for_guild(self.id)
        obbs = []

        for webhook in webhooks:
            obbs.append(self._bot.state.make_webhook(webhook))

        return obbs

    async def delete_webhook(self, webhook: 'dt_webhook.Webhook'):
        """
        Deletes a webhook in this guild.

        :param webhook: The webhook to delete.
        """
        if not self.me.guild_permissions.manage_webhooks:
            raise PermissionsError("manage_webhooks")

        await self._bot.http.delete_webhook(webhook.id)

    async def add_roles(self, member: 'dt_member.Member', *roles: typing.List['role.Role']):
        """
        Adds roles to a member.

        This will wait until the gateway returns the GUILD_MEMBER_UPDATE with the new role list for the member before
        returning.

        .. code:: python

            roles = filter(lambda r: "Mod" in r.name, guild.roles)
            await guild.add_roles(member, *roles)

        :param member: The member to add roles to.
        :param roles: The roles to add.
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        # Ensure we can add all of these roles.
        for _r in roles:
            if _r >= self.me.top_role:
                raise HierachyError(
                    "Cannot add role {} - it has a higher or equal position to our top role".format(_r.name)
                )

        async def _listener(before, after: dt_member.Member):
            if after.id != member.id:
                return False

            if not all(role in after.roles for role in roles):
                return False

            return True

        role_ids = set([_r.id for _r in member.roles] + [_r.id for _r in roles])
        listener = await curio.spawn(self._bot.wait_for("member_update", _listener))

        try:
            await self._bot.http.modify_member_roles(self.id, member.id, role_ids)
        except:
            await listener.cancel()
            raise

        # Now wait for the event to happen on the gateway.
        await listener.join()

        return member

    async def remove_roles(self, member: 'dt_member.Member', *roles: 'typing.List[role.Role]'):
        """
        Removes roles from a member.

        This will wait until the gateway fires a GUILD_MEMBER_UPDATE.

        :param member: The member to remove roles from.
        :param roles: The roles to add.
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        for _r in roles:
            if _r >= self.me.top_role:
                raise HierachyError(
                    "Cannot remove role {} - it has a higher or equal position to our top role".format(_r.name)
                )

        async def _listener(before, after: dt_member.Member):
            if after.id != member.id:
                return False

            if not all(role not in after.roles for role in roles):
                return False

            return True

        # Calculate the roles to keep.
        to_keep = set(member.roles) - set(roles)

        role_ids = set([_r.id for _r in to_keep])
        listener = await curio.spawn(self._bot.wait_for("member_update", _listener))

        try:
            await self._bot.http.modify_member_roles(self.id, member.id, role_ids)
        except:
            await listener.cancel()
            raise
        await listener.join()

        return member

    async def change_nickname(self, member: 'dt_member.Member', new_nickname: str):
        """
        Changes the nickname of a member.

        :param member: The member to change the nickname of.
        :param new_nickname: The new nickname.
        """
        me = False
        if member == self.me:
            me = True
            if not self.me.guild_permissions.change_nickname:
                raise PermissionsError("change_nickname")
        else:
            if not self.me.guild_permissions.manage_nicknames:
                raise PermissionsError("manage_nicknames")

        if member.top_role >= self.me.top_role and member != self.me:
            raise HierachyError("Top role is equal to or lower than victim's top role")

        if new_nickname is not None and len(new_nickname) > 32:
            raise ValueError("Nicknames cannot be longer than 32 characters")

        coro = self._bot.http.change_nickname(self.id, new_nickname,
                                              member_id=member.id, me=me)

        async def _listener(before, after):
            return after.guild == self and after.id == member.id

        listener = await curio.spawn(self._bot.wait_for("member_update", _listener))  # type: curio.Task
        try:
            await coro
        except:
            await listener.cancel()
            raise
        await listener.join()

        return member

    async def change_role_positions(self, roles: 'typing.Union[typing.Dict[role.Role, int], '
                                                 'typing.List[typing.Tuple[role.Role, int]]]'):
        """
        Changes the positions of a mapping of roles.

        :param roles: A dict or iterable of two-item tuples of new roles that is in the format of (role, position).
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        if isinstance(roles, dict):
            roles = roles.items()

        to_send = [(str(r.id), new_position) for (r, new_position) in roles]
        await self._bot.http.change_roles_position(to_send)

    async def change_voice_state(self, member: 'dt_member.Member', *,
                                 deaf: bool = None, mute: bool = None):
        """
        Changes the voice state of a member.

        :param member: The member to change the voice state of.
        :param deaf: Should this member be deafened?
        :param mute: Should this member be muted?
        """
        if member.voice is None:
            raise CuriousError("Cannot change voice state of member not in voice")

        await self._bot.http.edit_member_voice_state(self.id, member.id, deaf=deaf, mute=mute)

    async def modify_guild(self, **kwargs):
        """
        Edits this guild.

        For a list of available arguments, see https://discordapp.com/developers/docs/resources/guild#modify-guild.
        """
        if not self.me.guild_permissions.manage_server:
            raise PermissionsError("manage_server")

        if "afk_channel" in kwargs:
            kwargs["afk_channel_id"] = kwargs.pop("afk_channel").id

        await self._bot.http.modify_guild(self.id, **kwargs)
        return self

    async def change_icon(self, icon_content: bytes):
        """
        Changes the icon for this guild.

        :param icon_content: The bytes that represent the icon of the guild.
        """
        if not self.me.guild_permissions.manage_server:
            raise PermissionsError("manage_server")

        image = base64ify(icon_content)
        await self._bot.http.modify_guild(self.id,
                                          icon_content=image)

    def upload_icon(self, path):
        """
        Uploads a new icon for the guild.

        :param path: A path-like object to use to upload.
        """
        with open(path, 'rb') as f:
            return self.change_icon(f.read())

    async def create_channel(self, **kwargs):
        """
        Creates a new channel in this guild.
        """
        if not self.me.guild_permissions.manage_channels:
            raise PermissionsError("manage_channels")

        if "type_" in kwargs:
            kwargs["type"] = kwargs["type_"]

        if "type" not in kwargs:
            kwargs["type"] = channel.ChannelType.TEXT

        channel_data = await self._bot.http.create_channel(self.id, **kwargs)
        channel_object = channel.Channel(client=self._bot, **channel_data)
        channel_object.guild = self

        return channel_object

    async def edit_channel(self, channel_object: 'channel.Channel', **kwargs):
        """
        Edits a channel in this guild.

        :param channel_object: The channel to edit.
        """
        if not channel_object.permissions(self.me).manage_channels:
            raise PermissionsError("manage_channels")

        if "type_" in kwargs:
            kwargs["type"] = kwargs["type_"]

        if "type" not in kwargs:
            kwargs["type"] = channel.ChannelType.TEXT

        await self._bot.http.edit_channel(channel_object.id, **kwargs)
        return channel_object

    async def delete_channel(self, channel: 'channel.Channel'):
        """
        Deletes a channel in this guild.

        :param channel: The channel to delete.
        """
        if not channel.permissions(self.me).manage_channels:
            raise PermissionsError("manaqe_channels")

        await self._bot.http.delete_channel(channel.id)
        return channel

    async def create_role(self, **kwargs):
        """
        Creates a new role in this guild.

        This does *not* edit the role in-place.
        :return: A new :class:`Role`.
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        role_obb = role.Role(client=self._bot, **(await self._bot.http.create_role(self.id)))
        self._roles[role_obb.id] = role_obb
        role_obb.guild = self
        return await self.edit_role(role_obb, **kwargs)

    async def edit_role(self, role: 'role.Role', *,
                        name: str = None, permissions: 'dt_permissions.Permissions' = None,
                        colour: int = None, position: int = None,
                        hoist: bool = None, mentionable: bool = None):
        """
        Edits a role.

        :param role: The role to edit.
        :param name: The name of the role.
        :param permissions: The permissions that the role has.
        :param colour: The colour of the role.
        :param position: The position in the sorting list that the role has.
        :param hoist: Is this role hoisted (shows separately in the role list)?
        :param mentionable: Is this mentionable by everyone?
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        if permissions is not None:
            if isinstance(permissions, dt_permissions.Permissions):
                permissions = permissions.bitfield

        await self._bot.http.edit_role(self.id, role.id,
                                       name=name, permissions=permissions, colour=colour, hoist=hoist,
                                       position=position, mentionable=mentionable)

        return role

    async def delete_role(self, role: 'role.Role'):
        """
        Deletes a role.

        :param role: The role to delete.
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        await self._bot.http.delete_role(self.id, role.id)

    async def get_widget_info(self) -> 'typing.Tuple[bool, typing.Union[None, channel.Channel]]':
        """
        Gets the widget info for the current guild.
        
        :return: A two-item tuple: If this widget is enabled, and the channel the widget has an invite for. 
        """
        info = await self._bot.http.get_widget_status(self.id)
        return info.get("enabled", False), self.channels.get(int(info.get("channel_id", 0)))

    async def edit_widget(self, *,
                          status: bool = None, channel: 'channel.Channel' = -1):
        """
        Edits the widget for this guild.
        
        :param status: The status of this widget: True or False.
        :param channel: The channel object to set the instant invite to.
        """
        if channel is None:
            channel_id = None
        elif channel == -1:
            channel_id = 0
        else:
            channel_id = channel.id

        await self._bot.http.edit_widget(self.id, enabled=status, channel_id=channel_id)



"""
Misc utilities shared throughout the library.
"""
import base64
import datetime

import collections
import imghdr
import inspect

import functools
import typing


class AsyncIteratorWrapper(collections.AsyncIterator):
    """
    Wraps a function so that it can be iterated over asynchronously.
    """

    def __init__(self, client, coro: typing.Awaitable[typing.List]):
        self.client = client
        self.coro = coro

        self.items = collections.deque()

        self._filled = False

    async def _fill(self):
        self.items.extend(await self.coro)
        self._filled = True

    async def __anext__(self):
        if not self._filled:
            await self._fill()

        try:
            return self.items.popleft()
        except IndexError:
            raise StopAsyncIteration


def base64ify(image_data: bytes):
    """
    Base64-ifys an image to send to discord.

    :param image_data: The data of the image to use.
    :return: A string containing the encoded image.
    """
    # Convert the avatar to base64.
    mimetype = imghdr.what(None, image_data)
    if not mimetype:
        raise ValueError("Invalid image type")

    b64_data = base64.b64encode(image_data).decode()
    return "data:{};base64,{}".format(mimetype, b64_data)


def to_datetime(timestamp: str) -> datetime.datetime:
    """
    Converts a Discord-formatted timestamp to a datetime object.

    :param timestamp: The timestamp to convert.
    :return: The :class:`datetime.datetime` object that corresponds to this datetime.
    """
    if timestamp is None:
        return

    if timestamp.endswith("+00:00"):
        timestamp = timestamp[:-6]

    try:
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")
    except ValueError:
        # wonky datetimes
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S")


def replace_quotes(item: str) -> str:
    """
    Replaces the quotes in a string, but only if they are un-escaped.

    :param item: The string to scan.
    :return: The string, with quotes replace.
    """
    # A list is used because it can be appended easily.
    final_str_arr = []

    for n, char in enumerate(item):
        # only operate if the previous char actually exists
        if n - 1 < 0:
            if char != '"':
                final_str_arr.append(char)

            continue

        # Complex quoting rules!
        # If it's a SINGLE backslash, don't append it.
        # If it's a double backslash, append it.
        if char == '\\':
            if item[n - 1] == "\\":
                # double backslash, append it
                final_str_arr.append(char)

            continue

        if char == '"':
            # check to see if it's escaped
            if item[n - 1] == '\\':
                # if the last char on final_str_arr is NOT a backslash, we want to keep it.
                if len(final_str_arr) > 0 and final_str_arr[-1] != '\\':
                    final_str_arr.append('"')

            continue

        # None of the above were hit, so add it anyway and continue.
        final_str_arr.append(char)

    return "".join(final_str_arr)


def _traverse_stack_for(t: type):
    """
    Traverses the stack for an object of type `t`.

    :param t: The type of the object.
    :return: The object, if found.
    """
    for fr in inspect.stack():
        frame = fr.frame
        try:
            locals = frame.locals
        except AttributeError:
            # idk
            continue
        else:
            for object in locals.values():
                if type(object) is t:
                    return object
        finally:
            # prevent reference cycles
            del fr


def _ad_getattr(self, key: str):
    try:
        return self[key]
    except KeyError as e:
        raise AttributeError(key) from e


attrdict = type("attrdict", (dict,), {
    "__getattr__": _ad_getattr,
    "__setattr__": dict.__setitem__,
    "__doc__": "A dict that allows attribute access as well as item access for "
               "keys."
})



"""
A special gateway client for Voice.
"""
import enum
import json
import logging
import queue
import socket
import threading
import zlib

import curio
import typing
from cuiows.client import WSClient
from cuiows.exc import WebsocketClosedError

from curious.core.gateway import Gateway


class VGatewayOp(enum.IntEnum):
    IDENTIFY = 0
    SELECT_PROTOCOL = 1
    READY = 2
    HEARTBEAT = 3
    SESSION_DESCRIPTION = 4
    SPEAKING = 5
    HELLO = 8


class _HeartbeatThread(threading.Thread):
    """
    A subclass of threading.Thread that sends heartbeats every <x> seconds.
    """

    def __init__(self, gw: 'VoiceGateway', heartbeat_interval: float, *args, **kwargs):
        # super() isn't used anywhere else so I'm scared to use it here
        threading.Thread.__init__(self, *args, **kwargs)
        self._gateway = gw
        self._hb_interval = heartbeat_interval

        # The event that tells us to stop heartbeating.
        self._stop_heartbeating = threading.Event()

        # This thread doesn't need to die before Python closes, so we can daemonize.
        # This means that Python won't wait on us to shutdown.
        self.daemon = True

    def run(self):
        """
        Heartbeats every <x> seconds.
        """
        # This uses `Event.wait()` to wait for the next heartbeat.
        # This is effectively a bootleg sleep(), but we can signal it at any time to stop.
        # Which is really cool!

        # Send the first heartbeat.
        self._send_heartbeat()
        while not self._stop_heartbeating.wait(self._hb_interval):
            self._send_heartbeat()

    def get_heartbeat(self):
        """
        :return: A heartbeat packet.
        """
        return {
            "op": int(VGatewayOp.HEARTBEAT),
            "d": self._gateway.sequence
        }

    def _send_heartbeat(self):
        hb = self.get_heartbeat()

        # Enqueue onto the sending queue.
        logger.debug("Voice session heartbeating with sequence {}".format(hb["d"]))
        self._gateway.send(hb)


logger = logging.getLogger("curious.voice")


class VoiceGateway(object):
    """
    A special type of websocket gateway that operates on the Voice connection.
    """

    def __init__(self, session_id: str, token: str, endpoint: str, user_id: str, guild_id: str):
        #: The current cuiows websocket object.
        self.websocket = None  # type: WSClient

        self._open = False

        self.session_id = session_id
        self.token = token
        self.user_id = user_id
        self.guild_id = guild_id
        self.sequence = 0

        #: The main gateway object.
        self.main_gateway = None  # type: Gateway

        #: The event queue.
        self._event_queue = queue.Queue()

        #: The sender task.
        self._sender_task = None  # type: curio.Task

        #: The current heartbeat thread.
        self._heartbeat_thread = None

        #: Voice server stuff
        self.endpoint = endpoint
        self.port = None  # type: int
        self.ssrc = None  # type: int
        self.secret_key = b""

        self._ready = curio.Event()
        self._got_secret_key = curio.Event()

    async def connect(self, url: str):
        """
        Connects the client to the websocket.

        This does not do ANY handshake steps.
        :param url: The URL to connect to.
        """
        logger.info("Opening connection to {}".format(url))
        self.websocket = await WSClient.connect(url)
        logger.info("Connected to gateway!")
        self._open = True
        return self.websocket

    # copied from main gateway
    async def _send_events(self):
        while self._open:
            try:
                next_item = await curio.abide(self._event_queue.get)
            except curio.CancelledError:
                self._event_queue.put_nowait(next_item)
                return
            if isinstance(next_item, dict):
                # this'll come back around in a sec
                self._send_json(next_item)
            else:
                logger.debug("Sending websocket data {}".format(next_item))
                await self._send(next_item)

    async def _send(self, data):
        """
        Sends some data.

        This will actually send the data down the websocket, unlike `send` which only pretends to.
        """
        try:
            await self.websocket.send(data)
        except WebsocketClosedError:
            await self._close()

    def _send_json(self, payload: dict):
        """
        Sends a JSON payload down the websocket.

        This is a private method - use :meth:`send` instead.

        :param payload: The payload to send.
        """
        data = json.dumps(payload)
        return self.send(data)

    def send(self, data: typing.Any):
        """
        Enqueues some data to be sent down the websocket.

        This does not send the data immediately - it is enqueued to be sent.
        """
        try:
            self._event_queue.put(data, block=False)
        except queue.Full as e:
            raise RuntimeError("Gateway queue is full - this should never happen!") from e

    # Sending methods
    def send_identify(self):
        payload = {
            "op": VGatewayOp.IDENTIFY,
            "d": {
                "server_id": str(self.guild_id),
                "user_id": str(self.user_id),
                "session_id": str(self.session_id),
                "token": str(self.token)
            }
        }
        self._send_json(payload)

    def send_select_protocol(self, ip: str, port: int):
        payload = {
            "op": VGatewayOp.SELECT_PROTOCOL,
            "d": {
                "protocol": "udp",
                "data": {
                    "address": ip,
                    "port": port,
                    "mode": "xsalsa20_poly1305"
                }
            }
        }

        self._send_json(payload)

    def send_speaking(self, speaking: bool=True):
        payload = {
            "op": VGatewayOp.SPEAKING,
            "d": {
                "speaking": speaking,
                "delay": 0
            }
        }

        self._send_json(payload)

    def _start_heartbeating(self, heartbeat_interval: float) -> threading.Thread:
        """
        Starts the heartbeat thread.
        :param heartbeat_interval: The number of seconds between each heartbeat.
        """
        if self._heartbeat_thread:
            self._heartbeat_thread._stop_heartbeating.set()

        t = _HeartbeatThread(self, heartbeat_interval)
        t.start()
        self._heartbeat_thread = t
        return t

    async def _close(self):
        if not self.websocket.closed:
            await self.websocket.close_now(code=1000, reason="Client disconnected")

        self._open = False
        self._heartbeat_thread._stop_heartbeating.set()
        await self._sender_task.cancel()
        # put something on the queue to kill the old task
        self._event_queue.put(None)
        del self._event_queue

    @classmethod
    async def from_gateway(cls, gw: Gateway, guild_id: int, channel_id: int) -> 'VoiceGateway':
        """
        Opens a new voice gateway connection from an existing Gateway connection.

        :param gw: The existing :class:`Gateway` to create from.
        :param guild_id: The guild ID we're opening to.
        :param channel_id: The channel ID we're connecting to.
        """
        # Send our VOICE_STATE_UPDATE to tell Discord to start connecting us to voice.
        await gw.send_voice_state_update(guild_id, channel_id)
        state = await gw.state.wait_for_voice_data(guild_id)
        user_id = gw.state._user.id

        # Look up the gateway's IP address.
        endpoint = state["endpoint"]
        endpoint, port = endpoint.split(":")
        endpoint = await curio.abide(socket.gethostbyname, endpoint)

        obb = cls(token=state["token"], session_id=state["session_id"],
                  endpoint=endpoint,
                  user_id=user_id, guild_id=guild_id)
        # Open our connection to the voice websocket.
        if port == "80":
            await obb.connect("ws://{}".format(state["endpoint"]))
        elif port == "443":
            await obb.connect("ws://{}".format(state["endpoint"]))
        obb._sender_task = await curio.spawn(obb._send_events())
        # Send our IDENTIFY.
        obb.send_identify()

        return obb

    async def next_event(self):
        """
        Gets the next event, in decoded form.
        """
        if not self._open:
            raise WebsocketClosedError(1006, reason="Connection lost")

        try:
            event = await self.websocket.poll()
        except WebsocketClosedError:
            # Close ourselves.
            await self._close()
            raise

        if isinstance(event, (bytes, bytearray)):
            # decompress the message
            event = zlib.decompress(event, 15, 10490000)
            event = event.decode("utf-8")

        if event is None:
            return

        event = json.loads(event)

        op = event.get("op")
        data = event.get("d")
        seq = event.get("seq")

        # Heartbeat logic, same as normal websocket
        if seq is not None:
            self.sequence = int(seq)

        # Switch based on operator.
        if op == VGatewayOp.HELLO:
            # Ignore these, they're useless for now.
            return

        elif op == VGatewayOp.READY:
            # Start heartbeating.
            heartbeat_interval = data.get("heartbeat_interval", 45000) / 1000.0
            logger.debug("Heartbeating every {} seconds.".format(heartbeat_interval))
            self._start_heartbeating(heartbeat_interval)
            # Set our `ssrc`, `port` and `modes`.
            self.ssrc = data.get("ssrc")
            self.port = data.get("port")

            await self._ready.set()

        elif op == VGatewayOp.SESSION_DESCRIPTION:
            # Extract the secret key.
            self.secret_key = data.get('secret_key')
            self.send_speaking()
            await self._got_secret_key.set()

        elif op == VGatewayOp.HEARTBEAT:
            # silence
            pass

        else:
            logger.warning("Unhandled event: {}".format(op))



"""
Special helpers for events.
"""


def event(name, scan: bool=True):
    """
    Marks a function as an event.

    :param name: The name of the event.
    :param scan: Should this event be handled in scans too?
    """

    def __innr(f):
        f.event = name
        f.scan = scan
        return f

    return __innr


class EventContext(object):
    """
    Represents a special context that are passed to events.

    :ivar client: The client instance that the event is currently connected to.
    :ivar shard_id: The shard ID that this event was sent on.
    """

    def __init__(self, cl, shard_id: int):
        self.client = cl
        self.shard_id = shard_id
        self.shard_count = cl.shard_count

    def change_status(self, *args, **kwargs):
        kwargs["shard_id"] = self.shard_id
        return self.client.change_status(*args, **kwargs)

    @property
    def gateway(self):
        return self.client._gateways[self.shard_id]



class CommandsError(Exception):
    pass


class CheckFailureError(Exception):
    def __init__(self, ctx, check):
        self.ctx = ctx
        self.check = check

    def __repr__(self):
        if isinstance(self.check, list):
            return "The checks for `{.name}` failed.".format(self.ctx)
        return "The check `{.__name__}` for `{.name}` failed.".format(self.check, self.ctx)

    __str__ = __repr__


class MissingArgumentError(Exception):
    def __init__(self, ctx, arg):
        self.ctx = ctx
        self.arg = arg

    def __repr__(self):
        return "Missing required argument `{}` in `{.name}`.".format(self.arg, self.ctx)

    __str__ = __repr__


class CommandInvokeError(Exception):
    def __init__(self, ctx):
        self.ctx = ctx

    def __repr__(self):
        return "Command {.name} failed to invoke with error `{}`.".format(self.ctx, self.__cause__)

    __str__ = __repr__


class ConversionFailedError(Exception):
    def __init__(self, ctx, arg: str, to_type: type):
        self.ctx = ctx
        self.arg = arg
        self.to_type = to_type

    def __repr__(self):
        return "Cannot convert `{}` to type `{.__name__}`.".format(self.arg, self.to_type)

    __str__ = __repr__



import copy
import typing

from curious.dataclasses.bases import Dataclass
from curious.dataclasses.permissions import Permissions
from curious.dataclasses import role as dt_role
from curious.dataclasses.status import Game, Status
from curious.dataclasses import voice_state as dt_vs
from curious.dataclasses import user as dt_user
from curious.dataclasses import guild
from curious.util import to_datetime


class Member(Dataclass):
    """
    A member represents somebody who is inside a guild.

    :ivar id: The ID of this member.
    """

    __slots__ = ("_roles", "joined_at", "nickname", "guild", "game", "_status", "voice",)

    def __init__(self, client, **kwargs):
        super().__init__(kwargs["user"]["id"], client)

        # make the user to use and cache
        self._bot.state.make_user(kwargs["user"])

        #: A dictionary of :class:`Role` this user has.
        self._roles = {}

        #: The date the user joined the guild.
        self.joined_at = to_datetime(kwargs.pop("joined_at", None))

        #: The member's current nickname.
        self.nickname = kwargs.pop("nick", None)

        #: The member's current :class:`Guild`.
        self.guild = None  # type: guild.Guild

        #: The current :class:`Game` this Member is playing.
        self.game = None  # type: Game

        #: The current :class:`Status` of this member.
        self._status = None  # type: Status

        #: The current :class:`VoiceState` of this member.
        self.voice = None  # type: dt_vs.VoiceState

    def __hash__(self):
        return hash(self.guild.id) + hash(self.user.id)

    def __eq__(self, other):
        if not isinstance(other, Member):
            return NotImplemented

        return other.guild == self.guild and other.user == self.user

    def _copy(self):
        """
        Copies a member object.
        """
        new_object = object.__new__(self.__class__)  # type: Member
        new_object._bot = self._bot

        new_object.id = self.id
        new_object._roles = self._roles.copy()
        new_object.joined_at = self.joined_at
        new_object.guild = self.guild
        new_object.game = copy.deepcopy(self.game)
        new_object._status = self._status
        new_object.nickname = self.nickname

        return new_object

    @property
    def user(self) -> 'dt_user.User':
        """
        :return: The underlying user for this member.
        """
        return self._bot.state._users[self.id]

    @property
    def name(self) -> str:
        """
        :return: The computed display name of this user.
        """
        return self.nickname if self.nickname else self.user.username

    @property
    def mention(self):
        if self.nickname:
            return "<@!{}>".format(self.id)

        return self.user.mention

    @property
    def status(self) -> Status:
        """
        :return: The current status of this member.
        :rtype: Status
        """
        return self._status

    @status.setter
    def status(self, value):
        if value is None:
            return

        if not isinstance(value, Status):
            value = Status(value)
        self._status = value

    @property
    def roles(self) -> 'typing.Iterable[dt_role.Role]':
        """
        :return: A list of roles this user has.
        """
        return self._roles.values()

    @property
    def colour(self) -> int:
        """
        :return: The computed colour of this user.
        :rtype: int
        """
        roles = sorted(self.roles, reverse=True)
        roles = filter(lambda role: role.colour, roles)
        try:
            return next(roles).colour
        except StopIteration:
            return 0

    @property
    def top_role(self) -> 'dt_role.Role':
        """
        :return: This member's top-most :class:`Role`.
        """
        return next(iter(sorted(self.roles, reverse=True)), self.guild.default_role)

    @property
    def guild_permissions(self):
        """
        :return: The calculated guild permissions for a member.
        :rtype: Permissions
        """
        if self == self.guild.owner:
            return Permissions.all()

        bitfield = 0
        for role in self.roles:
            bitfield |= role.permissions.bitfield

        permissions = Permissions(bitfield)
        if permissions.administrator:
            return Permissions.all()

        return permissions

    # Member methods.
    def send(self, content: str, *args, **kwargs):
        return self.user.send(content, *args, **kwargs)

    def ban(self, delete_message_days: int=7):
        """
        Bans this member from the guild.

        :param delete_message_days: The number of days of messages to delete.
        """
        return self.guild.ban(self, delete_message_days=delete_message_days)

    def kick(self):
        """
        Kicks this member from the guild.
        """
        return self.guild.kick(self)

    def add_roles(self, *roles: 'typing.Iterable[dt_role.Role]'):
        """
        Adds roles to this member.

        For more documentation, see :meth:`Guild.add_roles`.

        :param roles: The list of roles to add.
        """
        return self.guild.add_roles(self, *roles)

    def remove_roles(self, *roles: 'typing.Iterable[dt_role.Role]'):
        """
        Removes roles from this member.

        For more documentation, see :meth:`Guild.remove_roles`.

        :param roles: The list of roles to remove.
        """
        return self.guild.remove_roles(self, *roles)

    def change_nickname(self, new_nickname: typing.Union[str, None]):
        """
        Changes the nickname of this member.

        :param new_nickname: The nickname to change to, None to remove the nickname.
        """
        return self.guild.change_nickname(self, new_nickname)



"""
The base class for a plugin.
"""
import functools

import typing

from curious.commands import context
from curious.core import client as cl


class Plugin(object):
    # mark as true to prevent being scanned
    _include_in_scan = False

    def __init__(self, bot: 'cl.Client'):
        self.bot = bot

    def __repr__(self):
        return "<Plugin name='{}'>".format(self.name, self.bot)

    @property
    def name(self):
        """
        Gets the name of this plugin.

        This is mostly for usage in subclasses to customize the name of the plugin from the default type name.
        """
        return self.__class__.__name__

    async def load(self):
        """
        Called just after the plugin has loaded, before any commands have been registered. Used for async loading.
        """

    async def unload(self):
        """
        Called just before the plugin is to be unloaded, after all commands have been removed.
        """

    async def plugin_check(self, ctx: 'context.Context'):
        """
        Added as a check for every command in this plugin.
        """
        return True

    def _scan_body(self) -> typing.Tuple[list, list]:
        """
        Scans the body of this type for events and commands.

        :return: Two lists, the first one containing events and the second one containing commands.
        """
        events = []
        commands = []

        for name, value in self.__class__.__dict__.items():
            if hasattr(value, "factory"):
                if getattr(value, "_subcommand", False):
                    # don't add subcommands
                    continue
                # this is set by the decorator to create a new command instance
                cmd = value.factory()
                commands.append(cmd)

            elif hasattr(value, "event"):
                def _wtf(v):
                    @functools.wraps(v)
                    def _event_wrapper(*args, **kwargs):
                        return v(self, *args, **kwargs)
                    return _event_wrapper

                events.append(_wtf(value))

        return events, commands

    @classmethod
    async def setup(cls, bot: 'cl.Client', *args, **kwargs):
        """
        Default setup function for a plugin.

        This will create a new instance of the class, then add it as a Plugin to the bot.
        """
        instance = cls(bot, *args, **kwargs)
        await instance.load()
        bot.add_plugin(instance)



import importlib
import inspect
import logging
import re
import sys
import traceback
from types import MappingProxyType

import curio
import multidict
import typing
from cuiows.exc import WebsocketClosedError
from curio.task import Task

from curious.commands import cmd, context, plugin
from curious.dataclasses.appinfo import AppInfo
from curious.dataclasses import guild as dt_guild
from curious.dataclasses.invite import Invite
from curious.dataclasses.message import Message
from curious.dataclasses.status import Game, Status
from curious.dataclasses.user import User, RelationshipUser
from curious.dataclasses.webhook import Webhook
from curious.dataclasses.widget import Widget
from curious.event import EventContext, event as ev_dec
from curious.exc import CuriousError
from curious.http.httpclient import HTTPClient
from curious.util import attrdict, base64ify

AUTOSHARD = object()


def split_message_content(content: str, delim: str = " ") -> typing.List[str]:
    """
    Splits a message into individual parts by `delim`, returning a list of strings.
    This method preserves quotes.

    :param content: The message content to split.
    :param delim: The delimiter to split on.
    :return: A list of items split
    """

    def replacer(m):
        return m.group(0).replace(delim, "\x00")

    parts = re.sub(r'".+?"', replacer, content).split()
    parts = [p.replace("\x00", " ") for p in parts]
    return parts


def _default_msg_func_factory(prefix: str):
    def __inner(bot: Client, message):
        matched = None
        if isinstance(prefix, str):
            match = message.content.startswith(prefix)
            if match:
                matched = prefix

        elif isinstance(prefix, list):
            for i in prefix:
                if message.content.startswith(i):
                    matched = i
                    break

        if not matched:
            return None

        tokens = split_message_content(message.content[len(matched):])
        command_word = tokens[0]

        return command_word, tokens[1:], matched

    __inner.prefix = prefix
    return __inner


class Client(object):
    """
    The main client class. This is used to interact with Discord.

    When creating a client object, you can either pass a token explicitly, or pass in in the :meth:`start` call or
    similar.

    .. code:: python

        bot = Client("'a'")  # pass explicitly
        bot.run("'b'")  # or pass to the run call.
    """

    def __init__(self, token: str = None, *,
                 enable_commands: bool = True,
                 command_prefix: typing.Union[str, list] = None,
                 message_check=None,
                 description: str = "The default curious description",
                 state_klass: type = None,
                 bot: bool=True, self_bot: bool=False):
        """
        :param token: The current token for this bot.
            This can be passed as None and can be initialized later.
            
        :param enable_commands: Should commands integration be enabled?
            If this is False, commands can still be registered etc, they just won't fire (the event won't appear).
        
        :param command_prefix: The command prefix for this bot.
        :param message_check: The message check function for this bot. 
            This should take two arguments, the client and message, and should return either None or a 3-item tuple:
              - The command word matched
              - The tokens after the command word
              - The prefix that was matched.
              
        :param description: The description of this bot for usage in the default help command.
            
        :param state_klass: The class to construct the connection state from.
        
        :param bot: If this client represents a bot user.
        :param self_bot: If this bot is a self bot (i.e only responds to the user).
        """
        #: The mapping of `shard_id -> gateway` objects.
        self._gateways = {}  # type: typing.Dict[int, Gateway]

        #: The number of shards this client has.
        self.shard_count = 0

        #: The token for the bot.
        self._token = token

        #: The current connection state for the bot.
        if state_klass is None:
            from curious.core.state import State
            state_klass = State
        self.state = state_klass(self)

        #: If this client is a bot.
        self.is_bot = bot

        if self.is_bot and self_bot is True:
            raise CuriousError("Cannot be a bot account and a self bot at the same time")

        #: If this is a self bot.
        self.self_bot = self_bot

        #: The current event storage.
        self.events = multidict.MultiDict()

        #: The current "temporary" listener storage.
        #: Temporary listeners are events that listen, and if they return True the listener is remove.
        #: They are used in the HTTP method by `wait=`, for example.
        self._temporary_listeners = multidict.MultiDict()

        #: The :class:`HTTPClient` used for this bot.
        self.http = None  # type: HTTPClient

        #: The cached gateway URL.
        self._gw_url = None  # type: str

        #: The application info for this bot.
        #: Instance of :class:`AppInfo`.
        self.application_info = None  # type: AppInfo

        #: The logger for this bot.
        self._logger = logging.getLogger("curious.client")

        if enable_commands:
            if not command_prefix and not message_check:
                raise TypeError("Must provide one of `command_prefix` or `message_check`")

            self._message_check = message_check or _default_msg_func_factory(command_prefix)
        else:
            self._message_check = lambda x: None

        #: The description of this bot.
        self.description = description

        #: The dictionary of command objects to use.
        self.commands = attrdict()

        #: The dictionary of plugins to use.
        self.plugins = {}

        self._plugin_modules = {}

        if enable_commands:
            # Add the handle_commands as a message_create event.
            self.add_event(self.handle_commands)

            from curious.commands.plugin_core import _Core
            self.add_plugin(_Core(self))

        self.scan_events()

    @property
    def user(self) -> User:
        """
        :return: The :class:`User` that this client is logged in as.
        """
        return self.state._user

    @property
    def guilds(self) -> 'typing.Mapping[int, dt_guild.Guild]':
        """
        :return: A mapping of :class:`Guild` that this client can see.
        """
        return self.state.guilds

    @property
    def invite_url(self):
        """
        :return: The invite URL for this bot.
        """
        return "https://discordapp.com/oauth2/authorize?client_id={}&scope=bot".format(self.application_info.client_id)

    async def get_gateway_url(self):
        if self._gw_url:
            return self._gw_url

        self._gw_url = await self.http.get_gateway_url()
        return self._gw_url

    async def get_shard_count(self):
        gw, shards = await self.http.get_shard_count()
        self._gw_url = gw

        return shards

    def guilds_for(self, shard_id: int) -> 'typing.Iterable[dt_guild.Guild]':
        """
        Gets the guilds for this shard.

        :param shard_id: The shard ID to get guilds from.
        :return: A list of :class:`Guild` that client can see on the specified shard.
        """
        return self.state.guilds_for_shard(shard_id)

    # Events
    def add_event(self, func, name: str = None):
        """
        Add an event to the internal registry of events.

        :param name: The event name to register under.
        :param func: The function to add.
        """
        if not inspect.iscoroutinefunction(func):
            raise TypeError("Event must be a coroutine function")

        if name is None:
            name = func.event

        self.events.add(name, func)

    def add_listener(self, name: str, func):
        """
        Adds a temporary listener.

        :param name: The name of the event to listen under.
        :param func: The callable to call.
        """
        if not inspect.iscoroutinefunction(func):
            raise TypeError("Listener must be a coroutine function")

        self._temporary_listeners.add(name, func)

    def remove_event(self, name: str, func):
        """
        Removes a function event.

        :param name: The name the event is registered under/.
        :param func: The function to remove.
        """
        a = self.events.getall(name)
        if func in a:
            a.remove(func)

        self.events.pop(name)
        for item in a:
            self.events.add(name, item)

    def event(self, name: str):
        """
        A convenience decorator to mark a function as an event.

        This will copy it to the events dictionary, where it will be used as an event later on.

        :param name: The name of the event.
        """

        def _inner(func):
            f = ev_dec(name)(func)
            self.add_event(func=f)

        return _inner

    def scan_events(self):
        """
        Scans this class for functions marked with an event decorator.
        """
        for _, item in inspect.getmembers(self, predicate=lambda x: hasattr(x, "event") and getattr(x, "scan", False)):
            self._logger.info("Registering event function {} for event {}".format(_, item.event))
            self.add_event(item)

    async def _error_wrapper(self, func, *args, **kwargs):
        try:
            await func(*args, **kwargs)
        except Exception as e:
            self._logger.exception("Unhandled exception in {}!".format(func.__name__))

    async def _wrap_context(self, ctx: 'context.Context'):
        """
        Wraps a context in a safety wrapper.

        This will dispatch `command_exception` when an error happens.
        """
        try:
            await ctx.invoke()
        except Exception as e:
            gw = self._gateways[ctx.event_context.shard_id]
            await self.fire_event("command_error", e, ctx=ctx, gateway=gw)

    async def _temporary_wrapper(self, event, listener, *args, **kwargs):
        try:
            result = await listener(*args, **kwargs)
        except Exception as e:
            self._logger.exception("Unhandled exception in {}!".format(listener.__name__))
            return
        if result is True:
            # Complex removal bullshit
            try:
                items = self._temporary_listeners.getall(event)
            except KeyError:
                # early removal done already, ignore
                return

            try:
                items.remove(listener)
            except ValueError:
                # race condition bullshit
                return
            # remove all keys
            self._temporary_listeners.pop(event)

            for i in items:
                # re-add all new items
                self._temporary_listeners.add(event, i)

    def remove_listener_early(self, event, listener):
        """
        Removes a listener early.

        :param event: The event to remove from.
        :param listener: The listener to remove.
        """
        all = self._temporary_listeners.getall(event)
        if listener in all:
            all.remove(listener)

        self._temporary_listeners.pop(event)
        for i in all:
            self._temporary_listeners.add(event, i)

    async def fire_event(self, event_name: str, *args, **kwargs) -> typing.List[Task]:
        """
        Fires an event to run.

        This will wrap the events in Tasks and return a list of them.

        :param event_name: The event name to fire.
        :return: A :class:`list` of :class:`curio.task.Task` representing the events.
        """
        gateway = kwargs.pop("gateway")

        if "ctx" not in kwargs:
            ctx = EventContext(self, gateway.shard_id)
        else:
            ctx = kwargs.pop("ctx")

        coros = self.events.getall(event_name, [])

        temporary_listeners = self._temporary_listeners.getall(event_name, [])

        if not coros and not temporary_listeners:
            return

        self._logger.debug(
            "Dispatching event {} to {} listeners"
            " on shard {}".format(event_name, len(coros) + len(temporary_listeners), gateway.shard_id)
        )

        tasks = []
        for event in coros.copy():
            tasks.append(await curio.spawn(self._error_wrapper(event, ctx, *args, **kwargs), daemon=True))

        for listener in temporary_listeners:
            tasks.append(await curio.spawn(self._temporary_wrapper(event_name, listener, ctx, *args, **kwargs),
                                           daemon=True))

        return tasks

    # commands
    @ev_dec("command_error", scan=False)
    async def default_command_error(self, ctx, e):
        """
        Default error handler.

        This is meant to be overriden - normally it will just print the traceback.
        """
        if len(self.events.getall("command_error")) >= 2:
            # remove ourselves
            self.remove_event("command_error", self.default_command_error)
            return

        traceback.print_exception(None, e, e.__traceback__)

    @ev_dec("message_create", scan=False)
    async def handle_commands(self, event_ctx: EventContext, message: Message):
        """
        Handles invokation of commands.

        This is added as an event during initialization.
        """
        # oddities
        if message.author is None:
            return

        user = message.author.user if hasattr(message.author, "user") else message.author

        if self.self_bot and user != self.user:
            return

        if not message.content:
            # Minor optimization - don't fire on empty messages.
            return

        # use the message check function
        _ = self._message_check(self, message)
        if inspect.isawaitable(_):
            _ = await _

        if not _:
            # not a command, do not bother
            return

        # func should return command_word and args for the command
        command_word, args, prefix = _
        command = self.get_command(command_word)
        if command is None:
            return

        # Create the context object that will be passed in.
        ctx = context.Context(self, command=command, message=message,
                              event_ctx=event_ctx)
        ctx.prefix = prefix
        ctx.name = command_word
        ctx.raw_args = args

        await curio.spawn(self._wrap_context(ctx))

    def add_command(self, command_name: str, command: 'cmd.Command'):
        """
        Adds a command to the internal registry of commands.

        :param command_name: The name of the command to add.
        :param command: The command object to add.
        """
        if command_name in self.commands:
            if not self.commands[command_name]._overridable:
                raise ValueError("Command {} already exists".format(command_name))
            else:
                self.commands.pop(command_name)

        self.commands[command_name] = command

    def add_plugin(self, plugin_class):
        """
        Adds a plugin to the bot.

        :param plugin_class: The plugin class to add.
        """
        events, commands = plugin_class._scan_body()

        for event in events:
            event.plugin = plugin_class
            self.events.add(event.event, event)

        for command in commands:
            # Bind the command to the plugin.
            command.instance = plugin_class
            # dont add any aliases
            self.add_command(command.name, command)

        self.plugins[plugin_class.name] = plugin_class

    async def load_plugins_from(self, import_name: str, *args, **kwargs):
        """
        Loads a plugin and adds it to the internal registry.

        :param import_name: The import name of the plugin to add (e.g `bot.plugins.moderation`).
        """
        module = importlib.import_module(import_name)
        mod = [module, []]

        # Locate all of the Plugin types.
        for name, member in inspect.getmembers(module):
            if not isinstance(member, type):
                # We only want to inspect instances of type (i.e classes).
                continue

            inherits = inspect.getmro(member)
            # remove `member` from the mro
            # this prevents us registering Plugin as a plugin
            inherits = [i for i in inherits if i != member]
            if plugin.Plugin not in inherits:
                # Only inspect instances of plugin.
                continue

            if not hasattr(member, "__dict__"):
                # don't be difficult
                continue

            # evil cpython type implementation detail abuse
            if member.__dict__.get("_include_in_scan", True) is False:
                # this won't show up for subclasses.
                # so they always have to be explicitly set
                continue

            # Assume it has a setup method on it.
            result = member.setup(self, *args, **kwargs)
            if inspect.isawaitable(result):
                result = await result

            # Add it to the list of plugins we need to destroy when unloading.
            # We add the type, not the instance, as the instances are destroyed separately.
            mod[1].append(member)

        if len(mod[1]) == 0:
            raise ValueError("Plugin contained no plugin classes (classes that inherit from Plugin)")

        self._plugin_modules[import_name] = mod

    async def unload_plugins_from(self, import_name: str):
        """
        Unloads plugins from the specified import name.

        This is the opposite to :meth:`load_plugins_from`.
        """
        mod, plugins = self._plugin_modules[import_name]
        # we can't do anything with the module currently because it's still in our locals scope
        # so we have to wait to delete it
        # i'm not sure this is how python imports work but better to be safe
        # (especially w/ custom importers)

        for plugin in plugins:
            # find all the Command instances with this plugin as the body
            for name, command in self.commands.copy().items():
                # isinstance() checks ensures that the type of the instance is this plugin
                if isinstance(command.instance, plugin):
                    # make sure the instance isn't lingering around
                    command.instance = None
                    # rip
                    self.commands.pop(name)

            # remove all events registered
            for name, event in self.events.copy().items():
                # not a plugin event
                if not hasattr(event, "plugin"):
                    continue
                if isinstance(event.plugin, plugin):
                    event.plugins = None
                    self.remove_event(name, event)

            # now that all commands are dead, call `unload()` on all of the instances
            for name, instance in self.plugins.copy().items():
                if isinstance(instance, plugin):
                    r = instance.unload()
                    if inspect.isawaitable(r):
                        await r

                    self.plugins.pop(name)

        # remove from sys.modules and our own registry to remove all references
        del sys.modules[import_name]
        del self._plugin_modules[import_name]
        del mod

    def get_commands_for(self, plugin: 'plugin.Plugin') -> typing.Generator['cmd.Command', None, None]:
        """
        Gets the commands for the specified plugin.

        :param plugin: The plugin instance to get commands of.
        :return: A list of :class:`Command`.
        """
        for command in self.commands.copy().values():
            if command.instance == plugin:
                yield command

    def get_command(self, command_name: str) -> 'typing.Union[cmd.Command, None]':
        """
        Gets a command object for the specified command name.

        :param command_name: The name of the command.
        :return: The command object if found, otherwise None.
        """

        def _f(cmd: cmd.Command):
            return cmd.can_be_invoked_by(command_name)

        f = filter(_f, self.commands.values())
        return next(f, None)

    # Gateway functions
    async def change_status(self, game: Game = None, status: Status = Status.ONLINE,
                            shard_id: int = 0):
        """
        Changes the bot's current status.

        :param game: The game object to use. None for no game.
        :param status: The new status. Must be a :class:`Status` object.
        :param shard_id: The shard to change your status on.
        """
        gateway = self._gateways[shard_id]
        return await gateway.send_status(game, status)

    async def wait_for(self, event_name: str, predicate: typing.Callable = None):
        """
        Wait for an event to happen in the gateway.

        You can specify a check to happen to check if this event is the one to return.
        When the check returns True, the listener is removed and the event data is returned.
        For example, to wait for a message with the content `Heck`:

        .. code:: python

            message = await client.wait_for("message_create", predicate=lambda m: m.content == "Heck")

        You can pass any function to this predicate. If this function takes an error, it will remove the listener,
        then raise into your code.

        .. code:: python

            async def _closure(message):
                if message.author.id != 66237334693085184:
                    return False

                if message.content == "sparkling water > tap water":
                    return True

                return False

            wrong = await client.wait_for("message_create", predicate=_closure)

        :param event_name: The name of the event to wait for.
        :param predicate: An optional check function to return.
        :return: The result of the event.
        """
        if predicate is None:
            predicate = lambda *args, **kwargs: True

        event = curio.Event()
        result = None
        _exc = None

        async def __event_listener_inner(ctx: EventContext, *args, **kwargs):
            try:
                is_result = predicate(*args, **kwargs)
                if inspect.isawaitable(is_result):
                    is_result = await is_result
            except Exception as e:
                # It is NOT the result we want.
                nonlocal _exc
                _exc = e
                await event.set()
                # Return True so this listener dies.
                return True
            else:
                if is_result:
                    # It is the result we want, so set the event.
                    await event.set()
                    # Then we store the result.
                    nonlocal result
                    result = args  # TODO: Figure out keyword arguments
                    return True

                return False

        self.add_listener(event_name, __event_listener_inner)
        # Wait on the event to be set.
        try:
            await event.wait()
        except curio.CancelledError:
            # remove the listener
            self.remove_listener_early(event_name, __event_listener_inner)
            raise
        # If it's an exception, raise the exception.
        if _exc is not None:
            raise _exc
        # Otherwise, return the event result.
        return result

    # HTTP Functions
    async def edit_profile(self, *,
                           username: str = None,
                           avatar: bytes = None):
        """
        Edits the profile of this bot.

        The user is **not** edited in-place - instead, you must wait for the `USER_UPDATE` event to be fired on the
        websocket.

        :param username: The new username of the bot.
        :param avatar: The bytes-like object that represents the new avatar you wish to use.
        """
        if username:
            if not 2 <= len(username) <= 32:
                raise ValueError("Username must be 2-32 characters")

        if avatar:
            avatar = base64ify(avatar)

        await self.http.edit_profile(username, avatar)

    async def edit_avatar(self, path: str):
        """
        A higher-level way to change your avatar.
        This allows you to provide a path to the avatar file instead of having to read it in manually.

        :param path: The path-like object to the avatar file.
        """
        with open(path, 'rb') as f:
            await self.edit_profile(avatar=f.read())

    async def get_user(self, user_id: int) -> User:
        """
        Gets a user by ID.

        :param user_id: The ID of the user to get.
        :return: A new User object.
        """
        try:
            return self.state._users[user_id]
        except KeyError:
            u = self.state.make_user(await self.http.get_user(user_id))
            # decache it if we need to
            self.state._check_decache_user(u.id)
            return u

    async def get_application(self, application_id: int) -> AppInfo:
        """
        Gets an application by ID.

        :param application_id: The client ID of the application to fetch.
        :return: A new :class:`AppInfo` object corresponding to the application.
        """
        data = await self.http.get_app_info(application_id=application_id)
        appinfo = AppInfo(self, **data)

        return appinfo

    async def get_webhook(self, webhook_id: int) -> Webhook:
        """
        Gets a webhook by ID.

        :param webhook_id: The ID of the webhook to get.
        :return: A new Webhook object.
        """
        return self.state.make_webhook(await self.http.get_webhook(webhook_id))

    async def get_invite(self, invite_code: str) -> Invite:
        """
        Gets an invite by code.

        :param invite_code: The invite code to get.
        :return: A new Invite object.
        """
        return Invite(self, **(await self.http.get_invite(invite_code)))

    async def get_widget(self, guild_id: int) -> Widget:
        """
        Gets a widget from a guild.
        
        :param guild_id: The ID of the guild to get the widget of. 
        :return: A :class:`Widget` object.
        """
        data = await self.http.get_widget_data(guild_id)
        return Widget(self, **data)

    # Utility functions
    async def connect(self, token: str = None, shard_id: int = 1):
        """
        Connects the bot to the gateway.

        This will NOT poll for events - only open a websocket connection!
        """
        from curious.core.gateway import Gateway

        if token:
            self._token = token

        if not self.http:
            self.http = HTTPClient(self._token, bot=self.is_bot)

        if not self.application_info and self.is_bot:
            self.application_info = AppInfo(self, **(await self.http.get_app_info(None)))

        gateway_url = await self.get_gateway_url()
        self._gateways[shard_id] = await Gateway.from_token(self._token, self.state, gateway_url,
                                                            shard_id=shard_id, shard_count=self.shard_count)
        await self._gateways[shard_id].websocket.wait_for_ready()

        return self

    async def poll(self, shard_id: int):
        """
        Polls the gateway for the next event.

        :param shard_id: The shard ID of the gateway to shard.
        """
        from curious.core.gateway import ReconnectWebsocket
        gw = self._gateways[shard_id]
        while True:
            try:
                await gw.next_event()
            except WebsocketClosedError as e:
                # Try and handle the close.
                if e.reason == "Client closed connection":
                    # internal
                    return

                if e.code in [1000, 4007] or gw.session_id is None:
                    self._logger.info("Shard {} disconnected with code {}, creating new session".format(shard_id,
                                                                                                        e.code))
                    self.state._reset(gw.shard_id)
                    await gw.reconnect(resume=False)
                elif e.code not in (4004, 4011):
                    # Try and RESUME.
                    self._logger.info("Shard {} disconnected with close code {}, reason {}, "
                                      "attempting a reconnect.".format(shard_id, e.code, e.reason))
                    await gw.reconnect(resume=True)
                else:
                    raise
            except ReconnectWebsocket:
                # We've been told to reconnect, try and RESUME.
                await gw.reconnect(resume=True)

    async def start(self, token: str = None, shards: int = 1):
        """
        Starts the gateway polling loop.

        This is a convenience method that polls on all the shards at once. It will **not** restart them automatically.
        """
        self._logger.info("Starting bot with {} shards.".format(shards))
        self.shard_count = shards
        tasks = []
        for shard_id in range(0, shards):
            await self.connect(token, shard_id=shard_id)
            tasks.append(await curio.spawn(self.poll(shard_id)))
            self._logger.info("Sleeping for 5 seconds between shard creation.")
            await curio.sleep(5)

        wait = curio.wait(tasks)

        results = []

        # Wait for the next task.
        while True:
            task = await wait.next_done()  # type: curio.Task
            if task is None:
                break

            try:
                result = await task.join()
            except Exception as e:
                result = e
            finally:
                results.append(result)

        return results

    async def start_autosharded(self, token: str = None):
        """
        Starts the bot with an automatically set number of shards.
        """
        if token:
            self._token = token

        if not self.http:
            # autosharded always means start autosharded
            self.http = HTTPClient(self._token, bot=True)

        shards = await self.get_shard_count()
        self.shard_count = shards
        await self.start(token, shards=shards)

    def run(self, token: str = None, shards: typing.Union[int, object] = 1):
        """
        Runs your bot with Curio with the monitor enabled.

        :param token: The token to run with.
        :param shards: The number of shards to run.
            If this is None, the bot will autoshard.
        """
        if not self.is_bot and shards != 1:
            raise CuriousError("Cannot start user bots in sharded mode")

        try:
            kernel = curio.Kernel(with_monitor=True, warn_if_task_blocks_for=5)
        except TypeError:
            # old vers of curio
            kernel = curio.Kernel(with_monitor=True)
        if shards == AUTOSHARD:
            coro = self.start_autosharded(token)
        else:
            coro = self.start(token, shards=shards)

        try:
            return kernel.run(coro=coro, shutdown=True)
        except (KeyboardInterrupt, EOFError):
            self._logger.info("C-c/C-d received, killing bot.")
            # Cleanup.
            coros = []
            for gateway in self._gateways.values():
                coros.append(gateway.websocket.close_now(1000, reason="Client closed connection"))
                coros.append(gateway._close())

            async def __cleanup():
                tasks = []
                for task in coros:
                    tasks.append(await curio.spawn(task))

                self._logger.info("Need to wait for {} task(s) to complete.".format(len(tasks)))

                # silence exceptions
                await curio.gather(tasks, return_exceptions=True)
                self._logger.info("Clean-up complete.")
                raise SystemExit()

            return kernel.run(coro=__cleanup())

    @classmethod
    def from_token(cls, token: str = None):
        """
        Starts a bot from a token object.

        :param token: The token to use for the bot.
        """
        bot = cls(token)
        return bot.run()
